"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import os
import re
import shutil
import subprocess
import tempfile
import textwrap
from pathlib import Path

from lxml import etree

from radish import after, before, given, then, when


@before.each_scenario()
def create_temporary_directory_for_radish_run(scenario):
    """Create the temporary directory for the radish run"""
    scenario.context.ctx_dir = tempfile.mkdtemp(prefix="radish-")
    scenario.context.features_dir = os.path.join(scenario.context.ctx_dir, "features")
    scenario.context.matching_config_dir = os.path.join(
        scenario.context.ctx_dir, "matcher-configs"
    )
    scenario.context.base_dir = os.path.join(scenario.context.ctx_dir, "radish")
    scenario.context.results_dir = os.path.join(scenario.context.ctx_dir, "results")
    scenario.context.failure_report_path = os.path.join(
        scenario.context.results_dir, "failure_report"
    )

    # setup failure gather hooks
    failure_gather_hooks = """
from radish import after
from radish.models.state import State

@after.each_step()
def gather_failure(step):
    if step.state is not State.FAILED:
        return

    with open(r"{}", "w+", encoding="utf-8") as failure_report_file:
        failure_report_file.write("{{}}\\n".format(step.failure_report.name))
        failure_report_file.write("{{}}\\n".format(step.failure_report.reason))
    """.format(
        scenario.context.failure_report_path
    )

    os.makedirs(scenario.context.features_dir)
    os.makedirs(scenario.context.matching_config_dir)
    os.makedirs(scenario.context.base_dir)
    os.makedirs(scenario.context.results_dir)

    with open(
        os.path.join(scenario.context.base_dir, "failure_report_hook.py"),
        "w+",
        encoding="utf-8",
    ) as failure_report_hook_file:
        failure_report_hook_file.write(textwrap.dedent(failure_gather_hooks))


@after.each_scenario()
def remove_temporary_directory_for_radish_run(scenario):
    """Remove the temporary directory for the radish run"""
    shutil.rmtree(scenario.context.ctx_dir)


@given("the Feature File {feature_filename:QuotedString}")
def create_feature_file(step, feature_filename):
    """Create the Feature File given in the doc string"""
    assert (
        step.doc_string is not None
    ), "Please provide a Feature in the Step doc string"

    feature_contents = step.doc_string

    feature_path = os.path.join(step.context.features_dir, feature_filename)
    with open(feature_path, "w+", encoding="utf-8") as feature_file:
        feature_file.write(feature_contents)


@given("the base dir module {module_filename:QuotedString}")
def create_base_dir_module(step, module_filename):
    """Create the base dir module given in the doc string"""
    assert (
        step.doc_string is not None
    ), "Please provide a Python Module in the Step doc string"

    module_contents = step.doc_string
    module_path = os.path.join(step.context.base_dir, module_filename)
    with open(module_path, "w+", encoding="utf-8") as module_file:
        module_file.write(module_contents)


@given("the Matcher Config File {matching_config_filename:QuotedString}")
def create_matcher_config_file(step, matching_config_filename):
    """Create the Matcher Config File given in the doc string"""
    assert (
        step.doc_string is not None
    ), "Please provide a Matcher Config in the Step doc string"

    matcher_config_contents = step.doc_string

    matching_config_path = os.path.join(
        step.context.matching_config_dir, matching_config_filename
    )
    with open(matching_config_path, "w+", encoding="utf-8") as matching_config_file:
        matching_config_file.write(matcher_config_contents)


@when("the {feature_filename:QuotedString} is run")
def run_feature_file(step, feature_filename):
    """Run the given Feature File"""
    feature_path = os.path.join(step.context.features_dir, feature_filename)
    radish_command = [
        "coverage",
        "run",
        "-p",
        "-m",
        "radish",
        "-b",
        step.context.base_dir,
        feature_path,
        "-t",
    ]
    proc = subprocess.Popen(
        radish_command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    stdout, _ = proc.communicate()

    step.context.exit_code = proc.returncode
    step.context.stdout = stdout
    step.context.command = radish_command


@when(
    "the {feature_filename:QuotedString} is run with the options {radish_options:QuotedString}"
)
def run_feature_file_with_options(step, feature_filename, radish_options):
    """Run the given Feature File"""
    feature_path = os.path.join(step.context.features_dir, feature_filename)

    # replace ctx_dir in radish_options
    radish_options = radish_options.format(ctx_dir=step.context.ctx_dir)

    radish_command = [
        "coverage",
        "run",
        "-p",
        "-m",
        "radish",
        "-b",
        step.context.base_dir,
        feature_path,
        "-t",
    ] + radish_options.split()
    proc = subprocess.Popen(
        radish_command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    stdout, _ = proc.communicate()

    step.context.exit_code = proc.returncode
    step.context.stdout = stdout
    step.context.command = radish_command


@then("the exit code should be {exit_code:int}")
def expect_exit_code(step, exit_code):
    """Expect the exit code to be a certain integer"""
    assert step.context.exit_code == exit_code, (
        "Actual exit code was: {}\n".format(step.context.exit_code)
        + "stdout from radish run: '{}':\n".format(" ".join(step.context.command))
        + step.context.stdout.decode("utf-8")
    )


def assert_output(actual_stdout, expected_stdout):
    """Assert that the captured stdout matches"""
    for actual_stdout_line, expected_stdout_line in zip(
        actual_stdout.splitlines(), expected_stdout.splitlines()
    ):
        assert re.match(
            "^" + expected_stdout_line + "$", actual_stdout_line
        ), "{!r} == {!r}".format(expected_stdout_line, actual_stdout_line)


@then("the output to match:")
def expect_output(step):
    """Expect the output to match the regex in the doc string"""
    assert (
        step.doc_string is not None
    ), "Please provide an output in the Step doc string"

    actual_stdout = step.context.stdout.decode("utf-8").replace("\r", "")
    # print(actual_stdout)
    assert_output(actual_stdout, step.doc_string.replace("\r", ""))


@then("the run should fail with a {exc_type_name:word}")
def expect_fail_with_exc(step, exc_type_name):
    """Expect the run failed with an exception of the given type and message from doc string"""
    assert step.context.exit_code != 0, +"stdout from radish run: '{}':\n".format(
        " ".join(step.context.command)
    ) + step.context.stdout.decode("utf-8")

    assert os.path.exists(step.context.failure_report_path), (
        "No Step Failure Report was written\n"
        + "stdout from radish run: '{}':\n".format(
            " ".join(step.context.command)
        )  # noqa
        + step.context.stdout.decode("utf-8")
    )

    with open(
        step.context.failure_report_path, encoding="utf-8"
    ) as failure_report_file:
        actual_exc_type_name = failure_report_file.readline().strip()
        actual_exc_reason = failure_report_file.readline().strip()

    assert (
        actual_exc_type_name == exc_type_name
    ), "Exception types don't match '{}' == '{}'".format(
        actual_exc_type_name, exc_type_name
    )

    if step.doc_string is not None:
        assert step.doc_string.strip() in actual_exc_reason, "Reasons don't match"


@then("the run should fail with")
def expect_fail(step):
    """Expect the run failed with an exception"""
    stdout = step.context.stdout.decode("utf-8").replace("\r", "")

    assert step.context.exit_code != 0, (
        "Actual exit code was: {}\n".format(step.context.exit_code)
        + "stdout from radish run: '{}':\n".format(" ".join(step.context.command))
        + stdout
    )

    match = step.doc_string in stdout
    assert match, (
        "Searched for:\n" + step.doc_string + "\n" + "within stdout:\n" + stdout
    )


@then("the XML file {xmlfile:word} validates against {xsdfile:word}")
def expect_xml_validates(step, xmlfile, xsdfile):
    """Expect that the given XML file validates against the given XSD file"""
    xmlfile_path = os.path.join(step.context.ctx_dir, xmlfile)
    xsdfile_path = Path(__file__).absolute().parent.parent / "assets" / xsdfile

    xmltree = etree.parse(xmlfile_path)
    xsdtree = etree.parse(str(xsdfile_path))
    xsd = etree.XMLSchema(xsdtree)

    xsd.assert_(xmltree)


@then(
    "the XML file {xmlfile:word} has the value "
    "{expected_value:QuotedString} at {xpath:QuotedString}"
)
def expect_value_at_xpath(step, xmlfile, expected_value, xpath):
    """Expect the given value at the given xpath in the given xml file"""
    xmlfile_path = os.path.join(step.context.ctx_dir, xmlfile)
    tree = etree.parse(xmlfile_path)

    elements = tree.xpath(xpath)

    assert len(elements) != 0, "Cannot find any elements at '{}'".format(xpath)

    assert (
        len(elements) == 1
    ), "Found multiple elements at '{}'. Expected exactly one.".format(xpath)

    actual_value = elements[0]

    assert (
        actual_value == expected_value
    ), "Actual value '{}' didn't match expected value '{}'".format(
        actual_value, expected_value
    )


@then("the XML file {xmlfile:word} has the content:")
def expect_xml_content(step, xmlfile):
    """Expect the XML file content to match the given Doc String"""
    assert (
        step.doc_string is not None
    ), "Please provide the content in the Step doc string"

    xmlfile_path = os.path.join(step.context.ctx_dir, xmlfile)

    with open(xmlfile_path, encoding="utf-8") as f:
        actual_content = f.read().replace("\r", "")

    assert_output(actual_content, step.doc_string.replace("\r", ""))


@when("the {matching_config_filename:QuotedString} is tested")
def run_matching_config(step, matching_config_filename):
    """Run the given Feature File"""
    run_matching_config_with_options(step, matching_config_filename, "")


@when(
    "the {matching_config_filename:QuotedString} "
    "is tested with the options {radish_options:QuotedString}"
)
def run_matching_config_with_options(step, matching_config_filename, radish_options):
    """Run the given Feature File"""
    matching_config_path = os.path.join(
        step.context.matching_config_dir, matching_config_filename
    )
    radish_command = [
        "coverage",
        "run",
        "-p",
        "-m",
        "radish.step_testing",
        "-b",
        step.context.base_dir,
        matching_config_path,
    ] + radish_options.split()
    proc = subprocess.Popen(
        radish_command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    stdout, _ = proc.communicate()

    step.context.exit_code = proc.returncode
    step.context.stdout = stdout
    step.context.command = radish_command


@when("the {feature_filename:QuotedString} is parsed")
def run_parser(step, feature_filename):
    """Run the given Feature File"""
    feature_path = os.path.join(step.context.features_dir, feature_filename)
    radish_command = ["coverage", "run", "-p", "-m", "radish.parser", feature_path]
    proc = subprocess.Popen(
        radish_command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    stdout, _ = proc.communicate()

    step.context.exit_code = proc.returncode
    step.context.stdout = stdout
    step.context.command = radish_command



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import json
import functools

import click

from radish.errors import RadishError
from radish.extensionregistry import extension
from radish.hookregistry import after
from radish.models import ScenarioOutline, ScenarioLoop, State


@extension
class CucumberJSONWriter:
    """Cucumber JSON writer radish extension"""

    OPTIONS = [
        click.Option(
            param_decls=("--cucumber-json", "cucumber_json"),
            is_flag=True,
            help=("Report run results in a cucumber JSON file"),
        ),
        click.Option(
            param_decls=("--cucumber-json-file", "cucumber_json_file"),
            default="cucumber.json",
            help=("Specify the file path for the cucumber JSON report"),
        ),
    ]

    @classmethod
    def load(cls, config):
        if config.cucumber_json:
            return cls(
                config.cucumber_json_file, config.tag_expression, config.scenario_ids
            )
        else:
            return None

    def __init__(self, cucumber_json_file, tag_expression, scenario_ids):
        after.all(order=500)(
            functools.partial(
                generate_cucumber_json, cucumber_json_file, tag_expression, scenario_ids
            )
        )


def generate_cucumber_json(cucumber_json_file, tag_expression, scenario_ids, features):
    """Generate a cucumber JSON report for the run"""
    if not features:
        raise RadishError("No features given to generate cucumber json file")

    ccjson = []
    for feature in features:
        if not feature.has_to_run(tag_expression, scenario_ids):
            continue

        feature_description = "\n".join(feature.description)
        feature_tags = []
        for tag in feature.tags:
            feature_tags.append({"name": "@" + tag.name, "line": tag.line})

        feature_json = {
            "uri": str(feature.path),
            "type": "feature",
            "keyword": "Feature",
            "id": str(feature.id),
            "name": feature.short_description,
            "line": feature.line,
            "description": feature_description,
            "tags": feature_tags,
            "elements": [],
        }

        for rule in feature.rules:
            for scenario in rule.scenarios:
                if not scenario.has_to_run(tag_expression, scenario_ids):
                    continue

                if isinstance(scenario, ScenarioOutline):
                    scenario_type = "scenario_outline"
                elif isinstance(scenario, ScenarioLoop):
                    scenario_type = "scenario_loop"
                else:
                    scenario_type = "scenario"

                scenario_tags = []
                for tag in scenario.tags:
                    scenario_tags.append({"name": "@" + tag.name, "line": tag.line})
                scenario_json = {
                    "keyword": "Scenario",
                    "type": scenario_type,
                    "id": str(scenario.id),
                    "name": scenario.short_description,
                    "line": scenario.line,
                    "description": "",
                    "steps": [],
                    "tags": scenario_tags,
                }

                for step in scenario.steps:
                    step_json = {
                        "keyword": step.keyword,
                        "name": step.text,
                        "line": step.line,
                        "result": {
                            "status": step.state.name.lower(),
                            "duration": step.duration().total_seconds() * 1e9,
                        },
                    }
                    if step.state is State.FAILED:
                        step_json["result"][
                            "error_message"
                        ] = step.failure_report.reason
                    if step.state is State.UNTESTED:
                        if step.starttime is None:
                            step_json["result"]["status"] = State.PENDING.name.lower()
                        else:
                            step_json["result"]["status"] = State.SKIPPED.name.lower()
                    if step.embeddings:
                        step_json["embeddings"] = step.embeddings
                    scenario_json["steps"].append(step_json)
                feature_json["elements"].append(scenario_json)
        ccjson.append(feature_json)

    with open(cucumber_json_file, "w+") as f:
        content = json.dumps(ccjson, indent=4, sort_keys=True)
        f.write(content)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
import sys
from pathlib import Path

sys.path.insert(0, Path().absolute().parent)

from radish import __version__  # noqa


# -- Project information -----------------------------------------------------

project = "radish"
copyright = "2019, Timo Furrer"
author = "Timo Furrer"

# The full version, including alpha/beta/rc tags
release = __version__


# -- General configuration ---------------------------------------------------

# The master toctree document.
master_doc = "index"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"
html_theme_options = {
    "logo": "radish-bdd-logo-trans-bg.png",
    "logo_name": "radish",
    "github_user": "radish-bdd",
    "github_repo": "radish",
    "github_button": True,
    "github_banner": True,
}

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/radish-bdd-logo.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from parse_type import TypeBuilder  # noqa

from radish.errors import RadishError


class ParseTypeRegistry:
    """The ``ParseTypeRegistry`` keeps track of all custom registered parse type extensions"""

    def __init__(self):
        self.types = {}

    def register(self, name, func, pattern=None):
        """Register a new parse type"""
        if name in self.types:
            raise RadishError(
                "Cannot register custom parse type with name {} because it already exists".format(
                    name
                )
            )

        if pattern is not None:
            func.pattern = pattern

        self.types[name] = func

    def create_decorator(self, context=None):
        """Create the custom_type decorator.

        The created decorator is injected into the given ``dict``-like ``context`` object.
        If the given ``context`` is ``None`` the Steps will be injected into ``globals()``.
        """
        if context is None:
            context = globals()

        def __create_decorator(decorator_name):
            def __decorator(name, pattern):
                def __wrapper(func):
                    self.register(name, func, pattern)
                    return func

                return __wrapper

            return __decorator

        decorator_name = "custom_type"
        context[decorator_name] = __create_decorator(decorator_name)
        return decorator_name


#: Holds a global instance of the ParseTypeRegistry which shall be used
#  by all modules implementing Custom Parse Types.
registry = ParseTypeRegistry()


def register_custom_type(**kwargs):
    """Register a custom type

    The arguments are expected to be in the form of: type-name=func.
    """
    for name, func in kwargs.items():
        registry.register(name, func)


__all__ = registry.create_decorator()



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import textwrap

import colorful as cf

from radish.extensionregistry import extension
from radish.hookregistry import after, before
from radish.models.state import State
from radish.terrain import world

#: Holds the amount of spaces to indent per block
INDENT_STEP = " " * 4


@extension
class DotsFormatter:
    @classmethod
    def load(cls, config):
        if config.formatter == "Dots":
            return cls()
        else:
            return None

    STATE_SYMBOLS = {
        State.PASSED: ".",
        State.PENDING: "P",
        State.UNTESTED: "U",
        State.SKIPPED: "S",
        State.FAILED: "F",
    }

    def __init__(self):
        self.failed_steps = []

        before.each_feature(is_formatter=True)(self.write_feature_header)
        after.each_feature(is_formatter=True)(
            lambda *args, **kwargs: print("", flush=True)
        )

        after.each_scenario(is_formatter=True)(self.write_dot_for_scenario)
        after.each_step(is_formatter=True)(self.remember_failed_steps)

        after.all(is_formatter=True)(self.write_endreport)

    def write_feature_header(self, feature):
        print(cf.bold_white(str(feature.path)) + ":", end=" ", flush=True)

    def write_dot_for_scenario(self, scenario):
        print(self.STATE_SYMBOLS[scenario.state], end="", flush=True)

    def remember_failed_steps(self, step):
        if step.state == State.FAILED:
            self.failed_steps.append(step)

    def write_endreport(self, features):
        if not self.failed_steps:
            return

        print()
        print(cf.bold_firebrick("Failures:"), flush=True)

        for step in self.failed_steps:
            print(
                "{}: {}".format(str(step.path), step.scenario.short_description),
                flush=True,
            )
            print(INDENT_STEP + "{} {}".format(step.keyword, step.text), flush=True)

            report = step.failure_report
            failure_report_indentation = INDENT_STEP + INDENT_STEP

            if world.config.with_traceback:
                failure_information = report.traceback
            else:
                failure_information = "{}: {}\n".format(report.name, report.reason)

            print(
                cf.firebrick(
                    textwrap.indent(failure_information, failure_report_indentation)
                ),
                flush=True,
            )



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import re
import sys
from io import StringIO

import click

from radish.errors import RadishError
from radish.extensionregistry import extension
from radish.hookregistry import before, after


@extension
class Coverage:
    """
    Extension for Python Code Coverage
    """

    OPTIONS = [
        click.Option(
            param_decls=("--with-coverage", "with_coverage"),
            is_flag=True,
            help="Enable Code Coverage",
        ),
        click.Option(
            param_decls=("--cover-package", "cover_packages"),
            multiple=True,
            help="Python Package name for which the coverage is measured and reported",
        ),
        click.Option(
            param_decls=("--cover-append", "cover_append"),
            is_flag=True,
            help="Append measured coverage data to previous collected data",
        ),
        click.Option(
            param_decls=("--cover-config-file", "cover_config_file"),
            default=".coveragerc",
            help="Path to a custom coverage configuration file",
        ),
        click.Option(
            param_decls=("--cover-branches", "cover_branches"),
            is_flag=True,
            help="Include branch coverage in the report",
        ),
        click.Option(
            param_decls=("--cover-erase", "cover_erase"),
            is_flag=True,
            help="Erase previously collected data",
        ),
        click.Option(
            param_decls=("--cover-min-percentage", "cover_min_percentage"),
            type=click.IntRange(0, 100),
            help="Fail if the provided minimum coverage percentage is not reached",
        ),
        click.Option(
            param_decls=("--cover-html", "cover_html"),
            help="Path to the directory where to store the HTML coverage report",
        ),
        click.Option(
            param_decls=("--cover-xml", "cover_xml"),
            help="Path to the directory where to store the XML coverage report",
        ),
    ]

    @classmethod
    def load(cls, config):
        if config.with_coverage:
            return cls(
                config.cover_packages,
                config.cover_append,
                config.cover_config_file,
                config.cover_branches,
                config.cover_erase,
                config.cover_min_percentage,
                config.cover_html,
                config.cover_xml,
            )
        else:
            return None

    def __init__(
        self,
        cover_packages,
        cover_append,
        cover_config_file,
        cover_branches,
        cover_erase,
        cover_min_percentage,
        cover_html,
        cover_xml,
    ):
        try:
            from coverage import Coverage  # noqa
        except ImportError:
            raise RadishError(
                "if you want to use the code coverage extension you have to "
                "'pip install radish-bdd[coverage]'"
            )

        self.cover_packages = cover_packages
        self.cover_append = cover_append
        self.cover_config_file = cover_config_file
        self.cover_branches = cover_branches
        self.cover_erase = cover_erase
        self.cover_min_percentage = cover_min_percentage
        self.cover_html = cover_html
        self.cover_xml = cover_xml

        before.all()(self.coverage_start)
        after.all()(self.coverage_stop)

        self.coverage = None
        self.modules_on_init = set(sys.modules.keys())

    def coverage_start(self, features):
        """
        Hook to start the coverage measurement
        """
        from coverage import Coverage

        # if no explicit modules are specified we just
        # use the ones loaded from radish's basedir.
        # During the plugin init the basedir modules are
        # not loaded yet, but they are during the start method.
        # Thus, we are safe to consider the difference between the
        # two for coverage measurement.
        if not self.cover_packages:
            source = list(set(sys.modules.keys()).difference(self.modules_on_init))
        else:
            source = self.cover_packages

        self.coverage = Coverage(
            source=source,
            config_file=self.cover_config_file,
            branch=self.cover_branches,
        )
        if self.cover_erase:
            self.coverage.combine()
            self.coverage.erase()

        if self.cover_append:
            self.coverage.load()
        self.coverage.start()

    def coverage_stop(self, features):
        """
        Stop the coverage measurement
        and create report
        """
        self.coverage.stop()
        self.coverage.combine()
        self.coverage.save()
        self.coverage.report(file=sys.stdout)

        if self.cover_html:
            self.coverage.html_report(directory=self.cover_html)

        if self.cover_xml:
            self.coverage.xml_report(outfile=self.cover_xml)

        if self.cover_min_percentage:
            report = StringIO()
            self.coverage.report(file=report)
            match = re.search(r"^TOTAL\s+(.*)$", report.getvalue(), re.MULTILINE)
            if not match:
                raise RadishError("Failed to find total percentage in coverage report")

            total_percentage = int(match.groups()[0].split()[-1][:-1])
            if total_percentage < int(self.cover_min_percentage):
                raise RadishError(
                    "Failed to reach minimum expected coverage of {0}% (reached: {1}%)".format(
                        self.cover_min_percentage, total_percentage
                    )
                )



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import itertools
import textwrap
from collections import Counter
from datetime import timedelta

import click
import colorful as cf

from radish.extensionregistry import extension
from radish.hookregistry import after, before
from radish.models import DefaultRule
from radish.models.state import State
from radish.terrain import world

#: Holds the amount of spaces to indent per block
INDENT_STEP = " " * 4

#: Holds the ANSI escape sequence for a line up jump
LINE_UP_JUMP = "\r\033[A\033[K"


@extension
class GherkinFormatter:
    OPTIONS = [
        click.Option(
            param_decls=("--no-step-rewrites", "no_step_rewrites"),
            is_flag=True,
            help=(
                "Turn off all Step rewrites. "
                "Steps are rewritten once they finished running. [GherkinFormatter]"
            ),
        )
    ]

    @classmethod
    def load(cls, config):
        if config.formatter == "Gherkin":
            return cls()
        else:
            return None

    def __init__(self):
        before.each_feature(is_formatter=True)(write_feature_header)
        after.each_feature(is_formatter=True)(write_feature_footer)

        before.each_rule(is_formatter=True)(write_rule_header)

        before.each_scenario(is_formatter=True)(write_scenario_header)
        after.each_scenario(is_formatter=True)(write_scenario_footer)

        before.each_step(is_formatter=True)(write_step_running)
        after.each_step(is_formatter=True)(write_step_result)

        after.all(is_formatter=True)(write_summary)


def write_tagline(tag, indentation=""):
    tagline = cf.deepSkyBlue3("@{name}".format(name=tag.name))
    print(indentation + tagline, flush=True)


def write_feature_header(feature):
    """Write the Feature Header to stdout

    The Feature Header will be printed in the form of:

    @tag-a
    @tag-b
    Feature: short description
        description line 1
        description line 2
    """
    # write Tags
    for tag in feature.tags:
        write_tagline(tag)

    # write Feature heading
    feature_heading = "{feature_keyword}: {short_description}".format(
        feature_keyword=cf.bold_white(feature.keyword),
        short_description=cf.white(feature.short_description),
    )
    print(feature_heading, flush=True)

    # write Feature description if available
    if feature.description:
        feature_description = "\n".join(
            INDENT_STEP + line for line in feature.description
        )
        print(feature_description + "\n", flush=True)

    # write Background if available
    if feature.background:
        background = "{background_keyword}: {short_description}".format(
            background_keyword=cf.bold_white(feature.background.keyword),
            short_description=cf.white(feature.background.short_description)
            if feature.background.short_description
            else "",
        )

        print(INDENT_STEP + background, flush=True)

        # TODO: write background steps
        for step in feature.background.steps:
            write_step(step, cf.deepSkyBlue3, indentation=INDENT_STEP + INDENT_STEP)

        print("", flush=True)


def write_feature_footer(feature):
    """Write the Feature Footer

    The Feature Footer is a blank line in case the Feature
    was empty.
    """
    if not feature.description and not feature.rules:
        print("", flush=True)


def write_rule_header(rule):
    """Write the Rule header

    The short description is only written if it's not a DefaultRule
    """
    if isinstance(rule, DefaultRule):
        return

    rule_heading = "{rule_keyword}: {short_description}".format(
        rule_keyword=cf.bold_white(rule.keyword),
        short_description=cf.white(rule.short_description),
    )

    print(INDENT_STEP + rule_heading + "\n", flush=True)


def write_scenario_header(scenario):
    """Write the Scenario header"""
    indentation_level = 1 if isinstance(scenario.rule, DefaultRule) else 2
    indentation = INDENT_STEP * indentation_level

    scenario_heading = "{scenario_keyword}: {short_description}".format(
        scenario_keyword=cf.bold_white(scenario.keyword),
        short_description=cf.white(scenario.short_description),
    )

    for tag in scenario.tags:
        write_tagline(tag, indentation)
    print(indentation + scenario_heading, flush=True)


def write_scenario_footer(scenario):
    """Write the Scenario footer"""
    print(flush=True)


def write_step_running(step):
    """Write the Step before it's running"""
    if not world.config.no_step_rewrites:
        write_step(step, cf.orange)


def write_step_result(step):
    """Write the Step after it's ran"""
    step_color_func = None
    if step.state == State.PASSED:
        step_color_func = cf.forestGreen
    elif step.state == State.FAILED:
        step_color_func = cf.firebrick
    elif step.state == State.PENDING:
        step_color_func = cf.orange
    else:
        step_color_func = cf.deepSkyBlue3

    if not world.config.no_ansi and not world.config.no_step_rewrites:
        # calculate how many line-ups are needed to rewrite the entire Step
        step_doc_string_lines = (
            step.doc_string.count("\n") + 2 if step.doc_string else 0
        )
        step_data_table_lines = len(step.data_table) if step.data_table else 0
        line_jumps = 1 + step_doc_string_lines + step_data_table_lines
        print(LINE_UP_JUMP * line_jumps, end="", flush=True)

    write_step(step, step_color_func)

    if step.failure_report:
        indentation_level = 2 if isinstance(step.rule, DefaultRule) else 3
        indentation = INDENT_STEP * indentation_level
        failure_report_indentation = indentation + INDENT_STEP
        report = step.failure_report

        if world.config.with_traceback:
            failure_information = report.traceback
        else:
            failure_information = "{}: {}\n".format(report.name, report.reason)

        print(
            step_color_func(
                textwrap.indent(failure_information, failure_report_indentation)
            ),
            end="",
            flush=True,
        )


def write_summary(features):
    """Write the end report after all Feature Files are ran"""
    feature_states = [f.state for f in features]
    features_line = "{} Feature{} ({})".format(
        len(feature_states),
        "s" if len(feature_states) != 1 else "",
        ", ".join(
            "{} {}".format(v, k.name.lower())
            for k, v in Counter(feature_states).items()
        ),
    )

    scenarios = []
    rules_scenarios = (rule.scenarios for feature in features for rule in feature.rules)
    for scenario in itertools.chain(*rules_scenarios):
        if hasattr(scenario, "examples"):
            scenarios.extend(scenario.examples)
        else:
            scenarios.append(scenario)

    scenarios_line = "{} Scenario{} ({})".format(
        len(scenarios),
        "s" if len(scenarios) != 1 else "",
        ", ".join(
            "{} {}".format(v, k.name.lower())
            for k, v in Counter(s.state for s in scenarios).items()
        ),
    )

    steps = [s for s in scenarios for s in s.steps]
    steps_line = "{} Step{} ({})".format(
        len(steps),
        "s" if len(steps) != 1 else "",
        ", ".join(
            "{} {}".format(v, k.name.lower())
            for k, v in Counter(s.state for s in steps).items()
        ),
    )

    print(features_line, flush=True)
    print(scenarios_line, flush=True)
    print(steps_line, flush=True)

    # remind about pending Steps
    pending_steps = [s for s in steps if s.state is State.PENDING]
    if pending_steps:
        pending_step_implementations = {s.step_impl for s in pending_steps}
        print(
            cf.orange(
                "You have {} pending Step Implementation{} affecting {} Step{}:".format(
                    cf.bold_orange(len(pending_step_implementations)),
                    "s" if len(pending_step_implementations) != 1 else "",
                    cf.bold_orange(len(pending_steps)),
                    "s" if len(pending_steps) != 1 else "",
                )
            )
        )
        for pending_step_implementation in pending_step_implementations:
            print(
                cf.orange(
                    "*  '{} {}' @ {}:{}".format(
                        cf.bold_orange(pending_step_implementation.keyword),
                        cf.bold_orange(pending_step_implementation.pattern),
                        cf.bold_orange(
                            pending_step_implementation.func.__code__.co_filename
                        ),
                        cf.bold_orange(
                            pending_step_implementation.func.__code__.co_firstlineno
                        ),
                    )
                )
            )
        print(cf.orange("Note: This may be the reason for potentially failed Steps!"))

    total_duration = sum((f.duration() for f in features), timedelta())
    timing_information = cf.deepSkyBlue3(
        "Run {marker} finished within {duration} seconds".format(
            marker=cf.bold_deepSkyBlue3(world.config.marker),
            duration=cf.bold_deepSkyBlue3(total_duration.total_seconds()),
        )
    )

    print(timing_information, flush=True)


def write_step(step, step_color_func, indentation=None):
    """Write a Step with the given color function"""
    if indentation is None:
        indentation_level = 2 if isinstance(step.rule, DefaultRule) else 3
        indentation = INDENT_STEP * indentation_level

    step_text = "{step_keyword} {text}".format(
        step_keyword=step_color_func(step.used_keyword), text=step_color_func(step.text)
    )

    print(indentation + step_text, flush=True)

    if step.doc_string is not None:
        doc_string_indentation = indentation + INDENT_STEP
        print(doc_string_indentation + cf.white('"""'), flush=True)
        print(
            cf.deepSkyBlue3(textwrap.indent(step.doc_string, doc_string_indentation)),
            end="",
            flush=True,
        )
        print(cf.white(doc_string_indentation + '"""'), flush=True)

    if step.data_table is not None:
        data_table_indentation = indentation + INDENT_STEP
        pretty_table = pretty_print_table(step.data_table, cf.white, cf.deepSkyBlue3)
        print(textwrap.indent(pretty_table, data_table_indentation), flush=True)


def pretty_print_table(table, bar_color_func, value_color_func):
    """Pretty-print the given Table"""
    column_widths = [max(len(str(col)) for col in row) for row in zip(*table)]

    colored_bar = bar_color_func("|")
    pretty_table = []
    for row in table:
        pretty_row = "{0} {1} {0}".format(
            colored_bar,
            bar_color_func(" | ").join(
                value_color_func("{1: <{0}}").format(column_widths[col_idx], col_value)
                for col_idx, col_value in enumerate(row)
            ),
        )
        pretty_table.append(pretty_row)

    return "\n".join(pretty_table)



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import json
import re
from pathlib import Path

from lark import Lark, UnexpectedInput

from radish.models import PreconditionTag
from radish.parser.errors import (
    RadishLanguageNotFound,
    RadishMisplacedBackground,
    RadishMissingFeatureShortDescription,
    RadishMissingRuleShortDescription,
    RadishMissingScenarioShortDescription,
    RadishMultipleBackgrounds,
    RadishPreconditionScenarioDoesNotExist,
    RadishPreconditionScenarioRecursion,
    RadishScenarioLoopInvalidIterationsValue,
    RadishScenarioLoopMissingIterations,
    RadishScenarioOutlineExamplesInconsistentCellCount,
    RadishScenarioOutlineExamplesMissingClosingVBar,
    RadishScenarioOutlineExamplesMissingOpeningVBar,
    RadishScenarioOutlineWithoutExamples,
    RadishStepDataTableMissingClosingVBar,
    RadishStepDocStringNotClosed,
    RadishStepDoesNotStartWithKeyword,
)
from radish.parser.transformer import RadishGherkinTransformer, Transformer


class LanguageSpec:
    """Represents a gherkin language specification"""

    def __init__(self, code, keywords):
        self.code = code
        self.keywords = keywords
        self.first_level_step_keywords = {
            keywords["Given"],
            keywords["When"],
            keywords["Then"],
        }

    def __str__(self):
        return self.code

    def __call__(self, terminal):
        try:
            terminal.pattern.value = self.keywords[terminal.pattern.value]
        except KeyError:
            pass


class FeatureFileParser:
    """Radish Feature File Parser responsible to parse a single Feature File"""

    def __init__(
        self,
        grammerfile: Path = None,
        ast_transformer: Transformer = RadishGherkinTransformer,
        resolve_preconditions: bool = True,
    ) -> None:
        if grammerfile is None:
            grammerfile = Path(__file__).parent / "grammer.g"

        self.grammerfile = grammerfile
        self.resolve_preconditions = resolve_preconditions

        if ast_transformer is not None:
            self._transformer = ast_transformer()
        else:
            self._transformer = None

        self._current_feature_id = 1

        self._parsers = {}

    def _get_parser(self, language_spec):
        """Get a parser and lazy create it if necessary"""
        try:
            return self._parsers[language_spec.code]
        except KeyError:
            parser = Lark.open(
                str(self.grammerfile),
                parser="lalr",
                transformer=self._transformer,
                edit_terminals=language_spec,
            )
            self._parsers[language_spec.code] = parser
            return parser

    def parse(self, featurefile: Path):
        """Parse the given Feature File"""
        ast = self.parse_file(featurefile, self._current_feature_id)
        self._current_feature_id += 1
        if self.resolve_preconditions:
            self._resolve_preconditions(featurefile.parent, ast, {ast.path: ast})
        return ast

    def parse_file(self, featurefile: Path, feature_id: int = 0):
        """Parse the given Feature File using the parser"""
        with open(str(featurefile), "r", encoding="utf-8") as featurefile_f:
            contents = featurefile_f.read()
            return self.parse_contents(featurefile, contents, feature_id)

    def parse_contents(
        self, featurefile_path: Path, featurefile_contents: str, feature_id: int = 0
    ):
        # evaluate the language for the Feature File content
        language_spec = self._detect_language(featurefile_contents)

        # prepare the transformer for the Feature File
        if self._transformer is not None:
            self._transformer.prepare(
                language_spec, featurefile_path, featurefile_contents, feature_id
            )

        # get a parser
        parser = self._get_parser(language_spec)

        try:
            ast = parser.parse(featurefile_contents)
        except UnexpectedInput as exc:
            exc_class = exc.match_examples(
                parser.parse,
                {
                    RadishMissingFeatureShortDescription: RadishMissingFeatureShortDescription.examples,  # noqa
                    RadishMissingRuleShortDescription: RadishMissingRuleShortDescription.examples,
                    RadishMissingScenarioShortDescription: RadishMissingScenarioShortDescription.examples,  # noqa
                    RadishMisplacedBackground: RadishMisplacedBackground.examples,
                    RadishStepDoesNotStartWithKeyword: RadishStepDoesNotStartWithKeyword.examples,
                    RadishStepDocStringNotClosed: RadishStepDocStringNotClosed.examples,
                    RadishScenarioOutlineWithoutExamples: RadishScenarioOutlineWithoutExamples.examples,  # noqa
                    RadishScenarioOutlineExamplesMissingOpeningVBar: RadishScenarioOutlineExamplesMissingOpeningVBar.examples,  # noqa
                    RadishMultipleBackgrounds: RadishMultipleBackgrounds.examples,
                    RadishStepDataTableMissingClosingVBar: RadishStepDataTableMissingClosingVBar.examples,  # noqa
                    RadishScenarioLoopMissingIterations: RadishScenarioLoopMissingIterations.examples,  # noqa
                    RadishScenarioLoopInvalidIterationsValue: RadishScenarioLoopInvalidIterationsValue.examples,  # noqa
                    RadishScenarioOutlineExamplesMissingClosingVBar: RadishScenarioOutlineExamplesMissingClosingVBar.examples,  # noqa
                    RadishScenarioOutlineExamplesInconsistentCellCount: RadishScenarioOutlineExamplesInconsistentCellCount.examples,  # noqa
                },
            )
            if not exc_class:
                raise
            raise exc_class(exc.get_context(featurefile_contents), exc.line, exc.column)
        return ast

    def _detect_language(self, featurefile_contents: str):
        """Detect the specified language in the first line of the Feature File

        If no language code is detected ``en`` is used.
        If an unknown language code is detected an error is raised.
        """

        def __get_language_spec(code):
            language_spec_path = (
                Path(__file__).parent / "languages" / "{}.json".format(code)
            )
            if not language_spec_path.exists():
                raise RadishLanguageNotFound(code)

            with open(
                str(language_spec_path), "r", encoding="utf-8"
            ) as language_spec_file:
                keywords = json.load(language_spec_file)

            return LanguageSpec(code, keywords)

        match = re.match(
            r"^#\s*language:\s*(?P<code>[a-zA-Z-]{2,})", featurefile_contents.lstrip()
        )
        language_code = match.groupdict()["code"] if match else "en"
        return __get_language_spec(language_code)

    def _resolve_preconditions(self, features_rootdir, ast, visited_features):
        for scenario in (s for rules in ast.rules for s in rules.scenarios):
            preconditions = []
            for precondition_tag in (
                t for t in scenario.tags if isinstance(t, PreconditionTag)
            ):
                precondition_path = features_rootdir / precondition_tag.feature_filename

                if precondition_path not in visited_features:
                    precondition_ast = self.parse_file(precondition_path)
                    visited_features[precondition_ast.path] = precondition_ast
                    self._resolve_preconditions(
                        features_rootdir, precondition_ast, visited_features
                    )
                else:
                    precondition_ast = visited_features[precondition_path]

                precondition_scenarios = (
                    s for rules in precondition_ast.rules for s in rules.scenarios
                )
                for precondition_scenario in precondition_scenarios:
                    if (
                        precondition_scenario.short_description
                        == precondition_tag.scenario_short_description
                    ):
                        break
                else:
                    raise RadishPreconditionScenarioDoesNotExist(
                        precondition_path,
                        precondition_tag.scenario_short_description,
                        (
                            s.short_description
                            for rules in ast.rules
                            for s in rules.scenarios
                        ),
                    )

                # check if the precondition leads to a recursion
                if (
                    precondition_scenario in preconditions
                    or precondition_scenario == scenario
                ):
                    raise RadishPreconditionScenarioRecursion(
                        scenario, precondition_scenario
                    )

                preconditions.append(precondition_scenario)

            # assign preconditions
            scenario.set_preconditions(preconditions)



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""


import threading

world = threading.local()
world.__doc__ = """Thread-local radish contex object

This object can be used to attach arbitrary data like
variables, functions and other objects which
can be accessed later in Step Implementations and Hooks.

However, it's preferred to use scoped contexts
like :attr:`radish.Step.context`, :attr:`Scenario.context`
or :attr:`Feature.context` for data.
"""


def pick(func):
    """Add the given function to the ``world`` object

    This can be used to easier access helper functions in Steps and Hooks.
    """
    setattr(world, func.__name__, func)
    return func


world.pick = pick



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""


def test_correct_exports_on_package_level():
    # THEN, can import top-level exports
    from radish import (  # noqa
        # Step Decorators
        given,
        when,
        then,
        step,
        # Hooks & Terrain
        world,
        before,
        after,
        # Step Pattern Types
        custom_type,
        register_custom_type,
        TypeBuilder,
        # The radish parser
        FeatureFileParser,
        # Models
        Feature,
        Rule,
        DefaultRule,
        Background,
        Scenario,
        ScenarioOutline,
        ScenarioLoop,
        Step,
        Tag,
        PreconditionTag,
        ConstantTag,
        # Exception Types
        RadishError,
        # The radish-bdd package version
        __version__,
    )



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from radish.models.timed import Timed
from radish.hookregistry import before, after


def record_starttime(timed_model: Timed):
    """Record the start time for the given Model"""
    timed_model.record_starttime()


def record_endtime(timed_model: Timed):
    """Record the end time for the given Model"""
    timed_model.record_endtime()


before.each_feature(always=True)(lambda f: record_starttime(f))
before.each_scenario(always=True)(lambda f: record_starttime(f))
before.each_step(always=True)(lambda f: record_starttime(f))
after.each_feature(always=True)(lambda f: record_endtime(f))
after.each_scenario(always=True)(lambda f: record_endtime(f))
after.each_step(always=True)(lambda f: record_endtime(f))



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import os
import sys
import time
from pathlib import Path
import traceback

import click
import colorful as cf

import radish.loader as loader
from radish.config import Config
from radish.errors import RadishError
from radish.extensionregistry import registry as extension_registry
from radish.hookregistry import registry as hook_registry
from radish.logger import enable_radish_debug_mode_click_hook, logger
from radish.parser import FeatureFileParser
from radish.runner import Runner
from radish.stepregistry import registry as step_registry
from radish.terrain import world

# NOTE(TF): somehow doctest with pytest imports this module multiple times ...
if "doctest" not in sys.modules:
    # load radish built-in extensions
    __SOURCE_DIR__ = Path(__file__).absolute().parent
    __BUILT_IN_EXTENSIONS__ = [
        __SOURCE_DIR__ / "extensions",
        __SOURCE_DIR__ / "formatters",
    ]
    loaded_built_in_extensions = loader.load_modules(__BUILT_IN_EXTENSIONS__)


def expand_basedirs(ctx, param, basedirs):
    """Expand the given basedirs setting

    A single basedirs can contain multiple basedir locations.
    The location must be split by a colon (:) on UNIX and
    with a semicolon (;) on Windows.
    """
    expanded_basedirs = []
    for unexpanded_basedirs in basedirs:
        separator = ";" if os.name == "nt" else ":"

        basedir_gen = (
            Path(os.path.expandvars(b)).expanduser()
            for b in unexpanded_basedirs.split(separator)
        )
        for basedir in basedir_gen:
            if not basedir.exists():
                raise OSError("The basedir '{}' does not exist.".format(basedir))

            expanded_basedirs.append(basedir)
    return expanded_basedirs


def expand_feature_files(ctx, param, feature_files):
    """Expand the given feature files

    Expanding directories recursively for Feature Files
    """
    expanded_feature_files = []
    for feature_file_location in (Path(f) for f in feature_files):
        if feature_file_location.is_dir():
            expanded_feature_files.extend(
                list(feature_file_location.glob("**/*.feature"))
            )
        else:
            expanded_feature_files.append(feature_file_location)

    return expanded_feature_files


def evaluate_user_data(ctx, param, userdata):
    """Evaluate the given user-defined data

    The user data can be in a ``key=value`` or just
    ``key`` format.
    """
    userdata_dict = {}
    for raw_data in userdata:
        key, *value = raw_data.split("=")
        userdata_dict[key] = "=".join(value) if value else True
    if userdata_dict:
        logger.debug("Evaluated user-defined data: %s", userdata_dict)
    return userdata_dict


class CommandWithExtensionOptions(click.Command):
    """Click Command to extend the given Options with the radish extension options"""

    def __init__(self, *args, **kwargs):
        radish_params = kwargs.pop("params", [])
        extension_options = extension_registry.get_extension_options()
        super().__init__(*args, params=radish_params + extension_options, **kwargs)


@click.command(cls=CommandWithExtensionOptions)
@click.version_option()
@click.help_option("--help", "-h")
@click.option(
    "--debug",
    "-d",
    "enable_debug_mode",
    is_flag=True,
    help="Enable debug mode for radish itself",
    callback=enable_radish_debug_mode_click_hook,
)
@click.option(
    "--basedir",
    "-b",
    "basedirs",
    multiple=True,
    default=("radish",),
    type=str,
    callback=expand_basedirs,
    help=(
        "Specify the location of the Step Implementations. "
        "One '-b' can contain multiple locations, split by a colon (:) [UNIX] "
        "or semicolon (;) [Windows]"
    ),
)
@click.option(
    "--early-exit",
    "-e",
    "early_exit",
    is_flag=True,
    help="Aborts the Runner immediately when the first Scenario failed",
)
@click.option(
    "--tags",
    "tags",
    help="Filter for Features and Scenarios matching this Tag Expression",
)
@click.option(
    "--scenarios",
    "-s",
    "scenario_ids",
    default="",
    callback=lambda _, __, ids: [int(x.strip()) for x in ids.split(",") if x.strip()],
    help="Filter for Scenarios with it's Id in this comma-separated list",
)
@click.option(
    "--shuffle",
    "shuffle_scenarios",
    is_flag=True,
    help="Shuffle the running order for the Scenarios within a Feature",
)
@click.option(
    "--wip",
    "wip_mode",
    is_flag=True,
    help=(
        "Run in WIP mode. In the WIP mode every Scenario has to fail."
        "It's best combined by marked WIP Scenarios with a '@wip'-Tag "
        "and use '--tags wip' in combination with the '--wip' flag"
    ),
)
@click.option(
    "--dry-run",
    "dry_run_mode",
    is_flag=True,
    help=(
        "Run in Dry Run mode. In the Dry Run mode Steps are matched with "
        "their implementation, but not run. "
        "Only formatter Hooks are called"
    ),
)
@click.option(
    "--debug-steps",
    "debug_steps_mode",
    is_flag=True,
    help=(
        "Run in Debug Steps mode. In the Debug Steps mode every Step is "
        "run in a Python debugger"
    ),
)
@click.option(
    "--user-data",
    "-u",
    multiple=True,
    callback=evaluate_user_data,
    help=(
        "Add arbitrary user-defined data to the 'radish.world' object. "
        "The data can be in a 'key=value' form or just a key"
    ),
)
@click.option(
    "--marker",
    "-m",
    "marker",
    default=str(int(time.time())),
    help=("Unique Marker for this Run. " "The Marker can be used in reports and logs"),
)
@click.option(
    "--formatter",
    "-f",
    "formatter",
    default="Gherkin",
    type=click.Choice(["Gherkin", "Dots"]),
    help="The formatter to use for this Run. ",
)
@click.option(
    "--no-ansi",
    "no_ansi",
    is_flag=True,
    help=(
        "Turn off all ANSI sequences (colors, line rewrites) ."
        "This option is mainly used by the formatters"
    ),
)
@click.option(
    "--with-traceback",
    "-t",
    "with_traceback",
    is_flag=True,
    help=(
        "Show the Traceback for failed Steps. "
        "This option is only used by the formatters"
    ),
)
@click.argument(
    "feature_files",
    nargs=-1,
    type=click.Path(exists=True),
    callback=expand_feature_files,
)
def cli(**kwargs):
    """radish - The root from red to green. BDD tooling for Python.

    Use radish to run your Feature File.

    Provide the Feature Files to run in FEATURE_FILES.
    """
    config = Config(kwargs)
    world.config = config

    # turn of ANSI colors if requested
    if config.no_ansi:
        cf.disable()

    logger.debug(
        "Loaded %d built-in extension modules: %s",
        len(loaded_built_in_extensions),
        ", ".join(str(e) for e in loaded_built_in_extensions),
    )
    logger.debug("Feature Files: %s", ", ".join(str(p) for p in config.feature_files))
    logger.debug("Basedirs: %s", ", ".join(str(d) for d in config.basedirs))

    logger.debug("Loading extensions")
    loaded_extensions = extension_registry.load_extensions(config)
    logger.debug(
        "Loaded %d extensions: %s",
        len(loaded_extensions),
        ", ".join(type(e).__name__ for e in loaded_extensions),
    )

    parser = FeatureFileParser()

    features = []
    for feature_file in config.feature_files:
        logger.debug("Parsing Feature File %s", feature_file)
        try:
            feature_ast = parser.parse(feature_file)
            if feature_ast:
                features.append(feature_ast)
        except RadishError as exc:
            print("", flush=True)
            print(
                "An error occured while parsing the Feature File {}:".format(
                    feature_file
                ),
                flush=True,
            )
            print(exc, flush=True)
            sys.exit(1)

    logger.debug("Loading all modules from the basedirs")
    try:
        loaded_modules = loader.load_modules(config.basedirs)
    except Exception as exc:
        print("", flush=True)
        print("An error occured while loading modules from the basedirs:", flush=True)
        print(exc, flush=True)
        if config.with_traceback:
            print("", flush=True)
            traceback.print_exc()
        sys.exit(1)

    logger.debug(
        "Loaded %d modules from the basedirs: %s",
        len(loaded_modules),
        ", ".join(str(m) for m in loaded_modules),
    )

    exit_status = 0
    try:
        runner = Runner(
            config, step_registry=step_registry, hook_registry=hook_registry
        )
        logger.debug(
            "Starting Runner WIP mode: %r, Dry-Run mode: %r",
            config.wip_mode,
            config.dry_run_mode,
        )
        success = runner.start(features)
        logger.debug("Finished Runner with status %s", success)

        exit_status = 0 if success else 1
    except RadishError as exc:
        print("", flush=True)
        print("An error occured while running the Feature Files:", flush=True)
        print(exc, flush=True)
        if config.with_traceback:
            print("", flush=True)
            traceback.print_exc()
        exit_status = 1

    sys.exit(exit_status)


if __name__ == "__main__":
    cli()



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import click

from radish.extensionregistry import extension
from radish.hookregistry import before, after


@extension
class SyslogWriter:
    """Syslog Writer radish extension

    This extension is only supported on
    systems where the Python standard library supports the system logger
    (syslog). For example, this extension works on UNIX and UNIX-like
    systems (Linux), but will not work on Windows.
    """

    OPTIONS = [
        click.Option(
            param_decls=("--with-syslog-markers", "with_syslog_markers"),
            is_flag=True,
            help="Log a marker for each Feature, Scenario and Step to the syslog",
        )
    ]

    @classmethod
    def load(cls, config):
        if config.with_syslog_markers:
            return cls(config.marker)
        else:
            return None

    def __init__(self, marker):
        self.marker = marker
        # import syslog only if the extension got loaded
        # but not if the module got loaded.
        import syslog  # noqa

        before.all()(self.syslog_writer_before_all)
        before.each_feature()(self.syslog_writer_before_each_feature)
        before.each_scenario()(self.syslog_writer_before_each_scenario)
        before.each_step()(self.syslog_writer_before_each_step)
        after.all()(self.syslog_writer_after_all)
        after.each_feature()(self.syslog_writer_after_each_feature)
        after.each_scenario()(self.syslog_writer_after_each_scenario)
        after.each_step()(self.syslog_writer_after_each_step)

    def syslog_writer_before_all(self, features):
        import syslog

        syslog.openlog("radish")
        syslog.syslog("begin run {}".format(self.marker))

    def syslog_writer_after_all(self, features):
        import syslog

        syslog.syslog("end run {}".format(self.marker))
        syslog.closelog()

    def syslog_writer_before_each_feature(self, feature):
        import syslog

        syslog.syslog(
            "begin feature {}:{} {}".format(
                self.marker, feature.id, feature.short_description
            )
        )

    def syslog_writer_after_each_feature(self, feature):
        import syslog

        syslog.syslog(
            "end feature {}:{} {}".format(
                self.marker, feature.id, feature.short_description
            )
        )

    def syslog_writer_before_each_scenario(self, scenario):
        import syslog

        syslog.syslog(
            "begin scenario {}:{}.{} {}".format(
                self.marker,
                scenario.feature.id,
                scenario.id,
                scenario.short_description,
            )
        )

    def syslog_writer_after_each_scenario(self, scenario):
        import syslog

        syslog.syslog(
            "end scenario {}:{}.{} {}".format(
                self.marker,
                scenario.feature.id,
                scenario.id,
                scenario.short_description,
            )
        )

    def syslog_writer_before_each_step(self, step):
        import syslog

        syslog.syslog(
            "begin step {}:{}.{}.{} {} {}".format(
                self.marker,
                step.feature.id,
                step.scenario.id,
                step.id,
                step.keyword,
                step.text,
            )
        )

    def syslog_writer_after_each_step(self, step):
        import syslog

        syslog.syslog(
            "{} step {}:{}.{}.{} {} {}".format(
                step.state.name.lower(),
                self.marker,
                step.feature.id,
                step.scenario.id,
                step.id,
                step.keyword,
                step.text,
            )
        )



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from unittest.mock import call, ANY

import pytest

from radish.errors import RadishError
from radish.hookregistry import HookRegistry
from radish.models import ScenarioLoop, ScenarioOutline
from radish.models.state import State
from radish.runner import Runner


@pytest.fixture(name="hook_registry", scope="function")
def setup_fake_hookregistry(mocker):
    """Setup a fake HookRegistry

    This fake HookRegistry can be used to assert called hooks.
    """
    fake_hookregistry = mocker.MagicMock(spec=HookRegistry)

    return fake_hookregistry


def test_start_run_without_features(hook_registry, default_config):
    """When starting the Runner without any Features it should pass and call the ``all`` hooks"""
    # given
    runner = Runner(default_config, None, hook_registry)

    # when
    status = runner.start([])

    # then
    assert status
    hook_registry.call.assert_has_calls(
        [call("all", "before", False, []), call("all", "after", False, [])]
    )


def test_start_run_should_iterate_all_given_features(
    hook_registry, default_config, mocker
):
    """All Features given to run should be iterated"""
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_feature = mocker.MagicMock()

    first_feature = mocker.MagicMock(name="First Feature")
    second_feature = mocker.MagicMock(name="Second Feature")

    # when
    runner.start([first_feature, second_feature])

    # given
    runner.run_feature.assert_has_calls([call(first_feature), call(second_feature)])


def test_should_only_run_feature_which_have_to_run(
    hook_registry, default_config, mocker
):
    """The Runner should only run features which need to be run"""
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_feature = mocker.MagicMock()

    first_feature = mocker.MagicMock(name="First Feature")
    first_feature.has_to_run.return_value = True
    second_feature = mocker.MagicMock(name="Second Feature")
    second_feature.has_to_run.return_value = False
    third_feature = mocker.MagicMock(name="Third Feature")
    third_feature.has_to_run.return_value = True

    # when
    runner.start([first_feature, second_feature, third_feature])

    # given
    runner.run_feature.assert_has_calls([call(first_feature), call(third_feature)])


def test_should_not_exit_for_failed_feature_if_early_exit_not_set(
    hook_registry, default_config, mocker
):
    """The Runner should not exit early if a Feature failes if the early exit flag is not set"""
    # given
    default_config.early_exit = False

    runner = Runner(default_config, None, hook_registry)
    runner.run_feature = mocker.MagicMock()
    runner.run_feature.side_effects = [State.PASSED, State.FAILED, State.PASSED]

    first_feature = mocker.MagicMock(name="First Feature")
    second_feature = mocker.MagicMock(name="Second Feature")
    third_feature = mocker.MagicMock(name="Third Feature")

    # when
    runner.start([first_feature, second_feature, third_feature])

    # given
    runner.run_feature.assert_has_calls(
        [call(first_feature), call(second_feature), call(third_feature)]
    )


def test_should_exit_for_failed_feature_if_early_exit_set(
    hook_registry, default_config, mocker
):
    """The Runner should exit early if a Feature failes if the early exit flag is set"""
    # given
    default_config.early_exit = True

    runner = Runner(default_config, None, hook_registry)
    runner.run_feature = mocker.MagicMock()
    runner.run_feature.side_effect = [State.PASSED, State.FAILED, State.PASSED]

    first_feature = mocker.MagicMock(name="First Feature")
    second_feature = mocker.MagicMock(name="Second Feature")
    third_feature = mocker.MagicMock(name="Third Feature")

    # when
    runner.start([first_feature, second_feature, third_feature])

    # given
    runner.run_feature.assert_has_calls([call(first_feature), call(second_feature)])


def test_should_return_good_status_if_all_features_passed(
    hook_registry, default_config, mocker
):
    """
    The Runner should return a good status when finished if all features passed in normal mode
    """
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_feature = mocker.MagicMock()
    runner.run_feature.side_effect = [State.PASSED, State.PASSED]

    first_feature = mocker.MagicMock(name="First Feature", state=State.PASSED)
    second_feature = mocker.MagicMock(name="Second Feature", state=State.PASSED)

    # when
    status = runner.start([first_feature, second_feature])

    # given
    assert status


def test_should_return_failed_status_if_not_all_features_passed(
    hook_registry, default_config, mocker
):
    """
    The Runner should return a failure status when finished
    if not all features passed in normal mode
    """
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_feature = mocker.MagicMock()
    runner.run_feature.side_effect = [State.PASSED, State.FAILED]

    first_feature = mocker.MagicMock(name="First Feature", state=State.PASSED)
    second_feature = mocker.MagicMock(name="Second Feature", state=State.FAILED)

    # when
    status = runner.start([first_feature, second_feature])

    # given
    assert not status


def test_should_return_good_status_if_all_features_failed_in_wip_mode(
    hook_registry, default_config, mocker
):
    """
    The Runner should return a good status when finished if all features failed in WIP mode
    """
    # given
    default_config.wip_mode = True

    runner = Runner(default_config, None, hook_registry)
    runner.run_feature = mocker.MagicMock()
    runner.run_feature.side_effect = [State.FAILED, State.FAILED]

    first_feature = mocker.MagicMock(name="First Feature", state=State.FAILED)
    second_feature = mocker.MagicMock(name="Second Feature", state=State.FAILED)

    # when
    status = runner.start([first_feature, second_feature])

    # given
    assert status


def test_should_return_failed_status_if_any_feature_passed_in_wip_mode(
    hook_registry, default_config, mocker
):
    """
    The Runner should return a failure status when finished
    if any feature passed in WIP mode
    """
    # given
    default_config.wip_mode = True

    runner = Runner(default_config, None, hook_registry)
    runner.run_feature = mocker.MagicMock()
    runner.run_feature.side_effect = [State.PASSED, State.FAILED]

    first_feature = mocker.MagicMock(name="First Feature", state=State.PASSED)
    second_feature = mocker.MagicMock(name="Second Feature", state=State.FAILED)

    # when
    status = runner.start([first_feature, second_feature])

    # given
    assert not status


def test_runner_should_call_hooks_when_running_a_feature(
    hook_registry, default_config, mocker
):
    """The Runner should call the ``each_feature`` hooks when running a Feature"""
    # given
    runner = Runner(default_config, None, hook_registry)
    feature_mock = mocker.MagicMock(name="Feature")

    # when
    runner.run_feature(feature_mock)

    # then
    hook_registry.call.assert_has_calls(
        [
            call("each_feature", "before", False, feature_mock),
            call("each_feature", "after", False, feature_mock),
        ]
    )


def test_runner_should_run_each_rule_in_a_feature(
    hook_registry, default_config, mocker
):
    """The Runner should run each Rule from a Feature"""
    # given
    runner = Runner(default_config, None, hook_registry)
    feature_mock = mocker.MagicMock(name="Feature")
    first_rule = mocker.MagicMock(name="First Rule")
    second_rule = mocker.MagicMock(name="Second Rule")
    feature_mock.rules = [first_rule, second_rule]
    runner.run_rule = mocker.MagicMock()

    # when
    runner.run_feature(feature_mock)

    # then
    runner.run_rule.assert_has_calls([call(first_rule), call(second_rule)])


def test_runner_should_only_run_rule_which_need_to_be_run(
    hook_registry, default_config, mocker
):
    """The Runner should run only the Rules which need to be run"""
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_rule = mocker.MagicMock()

    feature_mock = mocker.MagicMock(name="Feature")
    first_rule = mocker.MagicMock(name="First Rule")
    first_rule.has_to_run.return_value = True
    second_rule = mocker.MagicMock(name="Second Rule")
    second_rule.has_to_run.return_value = False
    third_rule = mocker.MagicMock(name="Third Rule")
    third_rule.has_to_run.return_value = True
    feature_mock.rules = [first_rule, second_rule, third_rule]

    # when
    runner.run_feature(feature_mock)

    # then
    runner.run_rule.assert_has_calls([call(first_rule), call(third_rule)])


def test_runner_should_not_exit_for_failed_rule_if_early_exit_flag_is_not_set(
    hook_registry, default_config, mocker
):
    """The Runner should not exit for a failed Rule if the early exit flag is not set"""
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_rule = mocker.MagicMock()
    runner.run_rule.side_effect = [State.PASSED, State.FAILED, State.PASSED]

    feature_mock = mocker.MagicMock(name="Feature")
    first_rule = mocker.MagicMock(name="First Rule")
    second_rule = mocker.MagicMock(name="Second Rule")
    third_rule = mocker.MagicMock(name="Third Rule")
    feature_mock.rules = [first_rule, second_rule, third_rule]

    # when
    runner.run_feature(feature_mock)

    # then
    runner.run_rule.assert_has_calls(
        [call(first_rule), call(second_rule), call(third_rule)]
    )


def test_runner_should_exit_for_failed_rule_if_early_exit_flag_is_set(
    hook_registry, default_config, mocker
):
    """The Runner should exit for a failed Rule if the early exit flag is set"""
    # given
    default_config.early_exit = True

    runner = Runner(default_config, None, hook_registry)
    runner.run_rule = mocker.MagicMock()
    runner.run_rule.side_effect = [State.PASSED, State.FAILED, State.PASSED]

    feature_mock = mocker.MagicMock(name="Feature")
    first_rule = mocker.MagicMock(name="First Rule")
    second_rule = mocker.MagicMock(name="Second Rule")
    third_rule = mocker.MagicMock(name="Third Rule")
    feature_mock.rules = [first_rule, second_rule, third_rule]

    # when
    runner.run_feature(feature_mock)

    # then
    runner.run_rule.assert_has_calls([call(first_rule), call(second_rule)])


def test_runner_should_call_hooks_when_running_a_rule(
    hook_registry, default_config, mocker
):
    """The Runner should call the ``each_rule`` hooks when running a Rule"""
    # given
    runner = Runner(default_config, None, hook_registry)
    rule_mock = mocker.MagicMock(name="Rule")

    # when
    runner.run_rule(rule_mock)

    # then
    hook_registry.call.assert_has_calls(
        [
            call("each_rule", "before", False, rule_mock),
            call("each_rule", "after", False, rule_mock),
        ]
    )


def test_runner_should_iterate_all_scenarios_when_running_a_rule(
    hook_registry, default_config, mocker
):
    """The Runner should iterate all Scenarios when running a Rule"""
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_scenario = mocker.MagicMock()

    rule_mock = mocker.MagicMock(name="Rule")
    first_scenario = mocker.MagicMock(name="First Scenario")
    second_scenario = mocker.MagicMock(name="Second Scenario")
    rule_mock.scenarios = [first_scenario, second_scenario]

    # when
    runner.run_rule(rule_mock)

    # then
    runner.run_scenario.assert_has_calls([call(first_scenario), call(second_scenario)])


def test_runner_should_only_run_scenario_which_need_to_be_run(
    hook_registry, default_config, mocker
):
    """The Runner should only run Scenarios which need to be run"""
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_scenario = mocker.MagicMock()

    rule_mock = mocker.MagicMock(name="Rule")
    first_scenario = mocker.MagicMock(name="First Scenario")
    first_scenario.has_to_run.return_value = True
    second_scenario = mocker.MagicMock(name="Second Scenario")
    second_scenario.has_to_run.return_value = False
    third_scenario = mocker.MagicMock(name="Third Scenario")
    third_scenario.has_to_run.return_value = True
    rule_mock.scenarios = [first_scenario, second_scenario, third_scenario]

    # when
    runner.run_rule(rule_mock)

    # then
    runner.run_scenario.assert_has_calls([call(first_scenario), call(third_scenario)])


def test_runner_should_not_exit_for_failed_scenario_if_early_exit_flag_is_not_set(
    hook_registry, default_config, mocker
):
    """The Runner should not exit for a failed Scenario if the early exit flag is not set"""
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_scenario = mocker.MagicMock()
    runner.run_scenario.side_effect = [State.PASSED, State.FAILED, State.PASSED]

    rule_mock = mocker.MagicMock(name="Rule")
    first_scenario = mocker.MagicMock(name="First Scenario")
    second_scenario = mocker.MagicMock(name="Second Scenario")
    third_scenario = mocker.MagicMock(name="Third Scenario")
    rule_mock.scenarios = [first_scenario, second_scenario, third_scenario]

    # when
    runner.run_rule(rule_mock)

    # then
    runner.run_scenario.assert_has_calls(
        [call(first_scenario), call(second_scenario), call(third_scenario)]
    )


def test_runner_should_exit_for_failed_scenario_if_early_exit_flag_is_set(
    hook_registry, default_config, mocker
):
    """The Runner should exit for a failed Scenario if the early exit flag is set"""
    # given
    default_config.early_exit = True

    runner = Runner(default_config, None, hook_registry)
    runner.run_scenario = mocker.MagicMock()
    runner.run_scenario.side_effect = [State.PASSED, State.FAILED, State.PASSED]

    rule_mock = mocker.MagicMock(name="Rule")
    first_scenario = mocker.MagicMock(name="First Scenario")
    second_scenario = mocker.MagicMock(name="Second Scenario")
    third_scenario = mocker.MagicMock(name="Third Scenario")
    rule_mock.scenarios = [first_scenario, second_scenario, third_scenario]

    # when
    runner.run_rule(rule_mock)

    # then
    runner.run_scenario.assert_has_calls([call(first_scenario), call(second_scenario)])


@pytest.mark.parametrize("scenario_container_type", [ScenarioLoop, ScenarioOutline])
def test_runner_should_run_scenario_loop_and_outline_as_scenario_container(
    scenario_container_type, hook_registry, default_config, mocker
):
    """The Runner should run a ScenarioLoop and ScenarioOutline as a Scenario Container"""
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_scenario_container = mocker.MagicMock()

    rule_mock = mocker.MagicMock(name="Rule")
    scenario = mocker.MagicMock(
        name=scenario_container_type.__name__, spec=scenario_container_type
    )
    rule_mock.scenarios = [scenario]

    # when
    runner.run_rule(rule_mock)

    # then
    runner.run_scenario_container.assert_has_calls([call(scenario)])


def test_runner_should_shuffle_scenarios_in_a_rule_if_shuffle_scenarios_flag_set(
    hook_registry, default_config, mocker
):
    """
    The Runner should shuffle the Scenarios within a Rule before
    running them if the shuffle Scenarios flag is set
    """
    # given
    default_config.shuffle_scenarios = True

    runner = Runner(default_config, None, hook_registry)
    runner.run_scenario = mocker.MagicMock()

    rule_mock = mocker.MagicMock(name="Rule")
    scenario = mocker.MagicMock(name="Scenario")
    rule_mock.scenarios = [scenario]

    mocker.patch("random.sample")

    # when
    runner.run_rule(rule_mock)

    # then
    import random

    random.sample.assert_called_once_with(rule_mock.scenarios, len(rule_mock.scenarios))


def test_runner_should_iterate_all_scenarios_when_running_a_scenario_container(
    hook_registry, default_config, mocker
):
    """The Runner should iterate all Scenarios when running a Scenario Container"""
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_scenario = mocker.MagicMock()

    scenario_container_mock = mocker.MagicMock(name="Scenario Container")
    first_scenario = mocker.MagicMock(name="First Scenario")
    second_scenario = mocker.MagicMock(name="Second Scenario")
    scenario_container_mock.examples = [first_scenario, second_scenario]

    # when
    runner.run_scenario_container(scenario_container_mock)

    # then
    runner.run_scenario.assert_has_calls([call(first_scenario), call(second_scenario)])


def test_runner_should_only_run_scenario_in_a_scenario_container_which_need_to_be_run(
    hook_registry, default_config, mocker
):
    """The Runner should only run Scenarios in a Scenario Container which need to be run"""
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_scenario = mocker.MagicMock()

    scenario_container_mock = mocker.MagicMock(name="Scenario Container")
    first_scenario = mocker.MagicMock(name="First Scenario")
    first_scenario.has_to_run.return_value = True
    second_scenario = mocker.MagicMock(name="Second Scenario")
    second_scenario.has_to_run.return_value = False
    third_scenario = mocker.MagicMock(name="Third Scenario")
    third_scenario.has_to_run.return_value = True
    scenario_container_mock.examples = [first_scenario, second_scenario, third_scenario]

    # when
    runner.run_scenario_container(scenario_container_mock)

    # then
    runner.run_scenario.assert_has_calls([call(first_scenario), call(third_scenario)])


def test_runner_should_not_exit_for_failed_scenario_in_scenario_container_if_early_exit_flag_is_not_set(  # noqa
    hook_registry, default_config, mocker
):
    """
    The Runner should not exit for a failed Scenario
    in Scenario Container if the early exit flag is not set
    """
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_scenario = mocker.MagicMock()
    runner.run_scenario.side_effect = [State.PASSED, State.FAILED, State.PASSED]

    scenario_container_mock = mocker.MagicMock(name="Scenario Container")
    first_scenario = mocker.MagicMock(name="First Scenario")
    second_scenario = mocker.MagicMock(name="Second Scenario")
    third_scenario = mocker.MagicMock(name="Third Scenario")
    scenario_container_mock.examples = [first_scenario, second_scenario, third_scenario]

    # when
    runner.run_scenario_container(scenario_container_mock)

    # then
    runner.run_scenario.assert_has_calls(
        [call(first_scenario), call(second_scenario), call(third_scenario)]
    )


def test_runner_should_exit_for_failed_scenario_in_scenario_container_if_early_exit_flag_is_set(
    hook_registry, default_config, mocker
):
    """
    The Runner should exit for a failed Scenario in
    Scenario Container if the early exit flag is set
    """
    # given
    default_config.early_exit = True

    runner = Runner(default_config, None, hook_registry)
    runner.run_scenario = mocker.MagicMock()
    runner.run_scenario.side_effect = [State.PASSED, State.FAILED, State.PASSED]

    scenario_container_mock = mocker.MagicMock(name="Scenario Container")
    first_scenario = mocker.MagicMock(name="First Scenario")
    second_scenario = mocker.MagicMock(name="Second Scenario")
    third_scenario = mocker.MagicMock(name="Third Scenario")
    scenario_container_mock.examples = [first_scenario, second_scenario, third_scenario]

    # when
    runner.run_scenario_container(scenario_container_mock)

    # then
    runner.run_scenario.assert_has_calls([call(first_scenario), call(second_scenario)])


def test_runner_should_shuffle_scenarios_in_a_scenario_container_if_shuffle_scenarios_flag_set(
    hook_registry, default_config, mocker
):
    """
    The Runner should shuffle the Scenarios within a Scenario Container before
    running them if the shuffle Scenarios flag is set
    """
    # given
    default_config.shuffle_scenarios = True

    runner = Runner(default_config, None, hook_registry)
    runner.run_scenario = mocker.MagicMock()

    scenario_container_mock = mocker.MagicMock(name="Rule")
    scenario = mocker.MagicMock(name="Scenario")
    scenario_container_mock.examples = [scenario]

    mocker.patch("random.sample")

    # when
    runner.run_scenario_container(scenario_container_mock)

    # then
    import random

    random.sample.assert_called_once_with(
        scenario_container_mock.examples, len(scenario_container_mock.examples)
    )


def test_runner_should_call_hooks_when_runner_a_scenario(
    hook_registry, default_config, mocker
):
    """The Runner should call the ``each_scenario`` hooks when running a Scenario"""
    # given
    runner = Runner(default_config, None, hook_registry)
    scenario_mock = mocker.MagicMock(name="Scenario")

    # when
    runner.run_scenario(scenario_mock)

    # then
    hook_registry.call.assert_has_calls(
        [
            call("each_scenario", "before", False, scenario_mock),
            call("each_scenario", "after", False, scenario_mock),
        ]
    )


def test_runner_should_iterate_all_steps_in_a_scenario(
    hook_registry, default_config, mocker
):
    """The Runner should iterate all Steps in a Scenario"""
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_step = mocker.MagicMock()
    runner.run_step.return_value = State.PASSED

    scenario_mock = mocker.MagicMock(name="Scenario")
    scenario_mock.background = None
    first_step = mocker.MagicMock(name="First Step")
    second_step = mocker.MagicMock(name="Second Step")
    scenario_mock.steps = [first_step, second_step]

    # when
    runner.run_scenario(scenario_mock)

    # then
    runner.run_step.assert_has_calls([call(first_step), call(second_step)])


def test_runner_should_only_not_run_steps_in_a_scenario_if_background_not_passed(
    hook_registry, default_config, mocker
):
    """
    The Runner should not run Steps in a Scenario
    if the Background is available and did not pass.
    """
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_step = mocker.MagicMock()
    runner.run_step.return_value = State.PASSED

    scenario_mock = mocker.MagicMock(name="Scenario")
    scenario_mock.background.state = State.FAILED
    first_step = mocker.MagicMock(name="First Step")
    scenario_mock.steps = [first_step]

    # when
    runner.run_scenario(scenario_mock)

    # then
    runner.run_step.assert_not_called()


def test_runner_should_only_run_steps_in_scenario_if_background_passed(
    hook_registry, default_config, mocker
):
    """
    The Runner should only run Steps in a Scenario
    if the Background is available and did pass.
    """
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_step = mocker.MagicMock()
    runner.run_step.return_value = State.PASSED

    scenario_mock = mocker.MagicMock(name="Scenario")
    scenario_mock.background.state = State.PASSED
    first_step = mocker.MagicMock(name="First Step")
    second_step = mocker.MagicMock(name="Second Step")
    scenario_mock.steps = [first_step, second_step]

    # when
    runner.run_scenario(scenario_mock)

    # then
    runner.run_step.assert_has_calls([call(first_step), call(second_step)])


def test_runner_should_stop_running_steps_after_first_failed(
    hook_registry, default_config, mocker
):
    """The Runner should stop running Steps after the first Step failed in normal mode"""
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_step = mocker.MagicMock()
    runner.run_step.side_effect = [State.PASSED, State.FAILED]

    scenario_mock = mocker.MagicMock(name="Scenario")
    scenario_mock.background = None
    first_step = mocker.MagicMock(name="First Step")
    second_step = mocker.MagicMock(name="Second Step")
    scenario_mock.steps = [first_step, second_step]

    # when
    runner.run_scenario(scenario_mock)

    # then
    runner.run_step.assert_has_calls([call(first_step)])


def test_runner_should_continue_running_steps_when_step_is_skipped_or_pending(
    hook_registry, default_config, mocker
):
    """The Runner should continue running Steps after a Step is skipped"""
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_step = mocker.MagicMock()
    runner.run_step.side_effect = [
        State.PASSED,
        State.SKIPPED,
        State.PENDING,
        State.PASSED,
    ]

    scenario_mock = mocker.MagicMock(name="Scenario")
    scenario_mock.background = None
    first_step = mocker.MagicMock(name="First Step")
    second_step = mocker.MagicMock(name="Second Step")
    third_step = mocker.MagicMock(name="Third Step")
    fourth_step = mocker.MagicMock(name="Fourth Step")
    scenario_mock.steps = [first_step, second_step, third_step, fourth_step]

    # when
    runner.run_scenario(scenario_mock)

    # then
    runner.run_step.assert_has_calls(
        [call(first_step), call(second_step), call(third_step), call(fourth_step)]
    )


def test_runner_should_abort_if_step_is_still_running_after_running_it(
    hook_registry, default_config, mocker
):
    """The Runner should abort if a Steps State is still running after its ran"""
    # given
    runner = Runner(default_config, None, hook_registry)
    runner.run_step = mocker.MagicMock()
    runner.run_step.side_effect = [State.PASSED, State.RUNNING, State.PASSED]

    scenario_mock = mocker.MagicMock(name="Scenario")
    scenario_mock.background = None
    first_step = mocker.MagicMock(name="First Step")
    second_step = mocker.MagicMock(name="Second Step")
    third_step = mocker.MagicMock(name="Third Step")
    scenario_mock.steps = [first_step, second_step, third_step]

    # then
    with pytest.raises(RadishError):
        # when
        runner.run_scenario(scenario_mock)

    # then
    runner.run_step.assert_has_calls([call(first_step), call(second_step)])


def test_runner_should_run_all_steps_even_when_failed_in_dry_run_mode(
    hook_registry, default_config, mocker
):
    """The Runner should runn all Steps even when one failed in the dry run mode"""
    # given
    default_config.dry_run_mode = True

    runner = Runner(default_config, None, hook_registry)
    runner.run_step = mocker.MagicMock()
    runner.run_step.side_effect = [State.FAILED, State.UNTESTED]

    scenario_mock = mocker.MagicMock(name="Scenario")
    scenario_mock.background = None
    first_step = mocker.MagicMock(name="First Step")
    second_step = mocker.MagicMock(name="Second Step")
    scenario_mock.steps = [first_step, second_step]

    # when
    runner.run_scenario(scenario_mock)

    # then
    runner.run_step.assert_has_calls([call(first_step), call(second_step)])


def test_runner_should_run_steps_from_a_background(
    hook_registry, default_config, mocker
):
    """The Runner should run all Steps from a Background before running the Scenario steps"""
    # given
    default_config.dry_run_mode = True

    runner = Runner(default_config, None, hook_registry)
    runner.run_step = mocker.MagicMock()
    runner.run_step.side_effect = [
        State.PASSED,
        State.PASSED,
        State.PASSED,
        State.PASSED,
    ]

    scenario_mock = mocker.MagicMock(name="Scenario")

    first_background_step = mocker.MagicMock(name="First Background Step")
    second_background_step = mocker.MagicMock(name="Second Background Step")
    scenario_mock.background.steps = [first_background_step, second_background_step]
    scenario_mock.background.state = State.PASSED

    first_step = mocker.MagicMock(name="First Step")
    second_step = mocker.MagicMock(name="Second Step")
    scenario_mock.steps = [first_step, second_step]

    # when
    runner.run_scenario(scenario_mock)

    # then
    runner.run_step.assert_has_calls(
        [
            call(first_background_step),
            call(second_background_step),
            call(first_step),
            call(second_step),
        ]
    )


def test_runner_should_call_hooks_when_running_a_step(
    hook_registry, default_config, mocker
):
    """The Runner should call the ``each_step`` hooks when running a Step"""
    # given
    runner = Runner(default_config, None, hook_registry)
    step_mock = mocker.MagicMock(name="Step")

    # when
    runner.run_step(step_mock)

    # then
    hook_registry.call.assert_has_calls(
        [
            call("each_step", "before", False, step_mock),
            call("each_step", "after", False, step_mock),
        ]
    )


def test_runner_should_run_step_after_being_matched_in_normal_mode(
    hook_registry, default_config, mocker
):
    """
    The Runner should run a Step after it's being
    successfully matched with a Step Implementation
    in the normal mode.
    """
    # given
    runner = Runner(default_config, None, hook_registry)
    step_mock = mocker.MagicMock(name="Step")

    mocker.patch("radish.runner.matcher")

    # when
    runner.run_step(step_mock)

    # then
    step_mock.run.assert_called_once_with(ANY)


def test_runner_should_debug_step_after_being_matched_in_debug_steps_mode(
    hook_registry, default_config, mocker
):
    """
    The Runner should debug a Step after it's being
    successfully matched with a Step Implementation
    in the debug steps mode.
    """
    # given
    default_config.debug_steps_mode = True

    runner = Runner(default_config, None, hook_registry)
    step_mock = mocker.MagicMock(name="Step")

    mocker.patch("radish.runner.matcher")

    # when
    runner.run_step(step_mock)

    # then
    step_mock.debug.assert_called_once_with(ANY)


def test_runner_should_not_run_nor_debug_step_after_being_matched_in_dry_run_mode(
    hook_registry, default_config, mocker
):
    """
    The Runner should not run nor debug a Step after it's being
    successfully matched with a Step Implementation in the dry run mode.
    """
    # given
    default_config.dry_run_mode = True

    runner = Runner(default_config, None, hook_registry)
    step_mock = mocker.MagicMock(name="Step")

    mocker.patch("radish.runner.matcher")

    # when
    runner.run_step(step_mock)

    # then
    step_mock.run.assert_not_called()
    step_mock.debug.assert_not_called()


def test_runner_should_fail_step_when_it_cannot_be_matched(
    hook_registry, default_config, mocker
):
    """The Runner should fail a Step when it cannot be matched with any Step Implementation"""
    runner = Runner(default_config, None, hook_registry)
    step_mock = mocker.MagicMock(name="Step")

    matcher_mock = mocker.patch("radish.runner.matcher")
    match_exc = RadishError("buuh!")
    matcher_mock.match_step.side_effect = match_exc

    # when
    runner.run_step(step_mock)

    # then
    step_mock.fail.assert_called_once_with(match_exc)



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from radish.models.state import State


class Rule:
    """Represents a single instance of a Gherkin Rule"""

    def __init__(
        self, keyword: str, short_description, path: str, line: int, scenarios
    ) -> None:
        self.keyword = keyword
        self.short_description = short_description
        self.path = path
        self.line = line
        self.scenarios = scenarios

        self.feature = None

    def __repr__(self) -> str:
        return "<Rule: '{short_description}' with {scenarios} Scenarios @ {path}:{line}>".format(
            short_description=self.short_description,
            scenarios=len(self.scenarios),
            path=self.path,
            line=self.line,
        )  # pragma: no cover

    def set_feature(self, feature):
        """Set the Feature for this Rule"""
        self.feature = feature
        for scenario in self.scenarios:
            scenario.set_feature(feature)

    def set_background(self, background):
        """Set the Background for all Scenarios in this Rule"""
        for scenario in self.scenarios:
            scenario.set_background(background)

    @property
    def state(self):
        """Read-only property to get the State for this Feature"""
        return State.report_state(s.state for s in self.scenarios)

    def get_all_tags(self):
        """Return all Tags for this Rule

        These Tags include the ones inherited from the Feature.
        """
        return self.feature.tags

    def has_to_run(self, tag_expression, scenario_ids):
        """Evaluate if this Rule has to run or not

        A Rule has to run of any of it's Scenarios has to be run.
        """
        return any(s.has_to_run(tag_expression, scenario_ids) for s in self.scenarios)


class DefaultRule(Rule):
    """Represents the default Rule which is used if no rule is specified"""

    def __init__(self, path: str, line: int, scenarios) -> None:
        super().__init__(None, None, path, line, scenarios)

    def __repr__(self) -> str:
        return "<DefaultRule: with {scenarios} Scenarios @ {path}:{line}>".format(
            scenarios=len(self.scenarios), path=self.path, line=self.line
        )  # pragma: no cover



"""
radish
~~~~~~

the root from red to green.  BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import copy

from radish.models.scenario import Scenario


class ScenarioOutline(Scenario):
    """Represents a single instance of a Gherkin Scenario Outline"""

    def __init__(
        self,
        scenario_id: int,
        keyword: str,
        short_description: str,
        tags,
        path: str,
        line: int,
        steps,
        examples_table,
    ) -> None:
        super().__init__(
            scenario_id, keyword, short_description, tags, path, line, steps
        )
        self.examples_table = examples_table
        self.examples = self._build_examples(self.examples_table)

    def __repr__(self) -> str:
        return (
            "<ScenarioOutline: '{short_description} with "
            "{examples} Examples @ {path}:{line}>".format(
                short_description=self.short_description,
                examples=len(self.examples_table),
                path=self.path,
                line=self.line,
            )
        )  # pragma: no cover

    def set_feature(self, feature):
        """Set the Feature for this Scenario"""
        super().set_feature(feature)
        for example in self.examples:
            example.set_feature(feature)
            example.keyword = feature.language_spec.keywords["Scenario"]

    def set_background(self, background):
        super().set_background(background)
        for example in self.examples:
            example.set_background(self.background)

    def set_rule(self, rule):
        """Set the Rule for this Scenario"""
        super().set_rule(rule)
        for example in self.examples:
            example.set_rule(rule)

    def _build_examples(self, examples_table):
        """Build the examples from the Examples Table"""
        examples = []
        for example_id, example_decl in enumerate(examples_table, start=1):
            # patch Steps from Scenario Outline for Examples
            steps = copy.deepcopy(self.steps)

            # Example description
            example_short_description_data = [
                "{}: {}".format(k, v) for k, v in example_decl.items()
            ]
            short_description = "{} [{}]".format(
                self.short_description, ", ".join(example_short_description_data)
            )

            for step in steps:
                for example_key, example_value in example_decl.items():
                    step.text = step.text.replace(
                        "<{}>".format(example_key), example_value
                    )

            example = Scenario(
                self.id + example_id,
                "Scenario",  # NOTE(TF): keyword will be patched in ``set_feature()``
                short_description,
                self.tags,
                self.path,
                self.line,  # FIXME(TF): use correct line number
                steps,
            )
            examples.append(example)

        return examples

    def has_to_run(self, tag_expression, scenario_ids):
        """Evaluate if this Scenario has to run or not

        The Scenario needs to be run if any of the following conditions is True:
        * No ``tag_expression`` is given
        * The ``tag_expression`` evaluates to True on the Scenario Tags
        * The ``tag_expression`` evaluates to True on the Feature Tags
        * No ``scenario_ids`` is given
        * The ``scenario_ids`` is given and contains this Scenarios Id
        """
        has_to_run_tag_expression = True
        has_to_run_scenario_ids = True

        if tag_expression:
            tag_names = [t.name for t in self.tags + self.feature.tags]
            has_to_run_tag_expression = tag_expression.evaluate(tag_names)

        if scenario_ids:
            has_to_run_scenario_ids = (
                bool({e.id for e in self.examples} & set(scenario_ids))
                or self.id in scenario_ids  # noqa
            )

        if tag_expression and scenario_ids:
            return has_to_run_tag_expression and has_to_run_scenario_ids
        elif tag_expression:
            return has_to_run_tag_expression
        elif scenario_ids:
            return has_to_run_scenario_ids
        else:
            return True



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import re

import pytest

from radish.stepregistry import StepImpl
from radish.matcher import match_step, RegexStepImplMatcher, ParseTypeStepImplMatcher
from radish.models import Step
from radish.errors import (
    StepImplementationNotFoundError,
    StepImplementationPatternNotSupported,
)


def test_matcher_should_raise_error_when_no_step_impl_found(mocker):
    """The matcher should raise an error when no Step Implementation is found"""
    # given
    registry_mock = mocker.MagicMock()
    registry_mock.step_implementations.return_value = []
    step = Step(1, "Given", "Given", "pattern", None, None, None, None)
    step.set_scenario(mocker.MagicMock(name="Scenario"))

    # then
    with pytest.raises(StepImplementationNotFoundError) as excinfo:
        # when
        match_step(step, registry_mock)

    # then
    assert excinfo.value.step == step


def test_matcher_should_raise_error_when_no_matcher_for_pattern_type(mocker):
    """
    The matcher should raise an error when no Matcher supports a Step Implementation Pattern Type
    """
    # given
    class NotSupportedPatternType:
        pass

    step_impl = StepImpl("Given", NotSupportedPatternType(), None)
    registry_mock = mocker.MagicMock()
    registry_mock.step_implementations.return_value = [step_impl]
    step = Step(1, "Given", "Given", "pattern", None, None, None, None)
    step.set_scenario(mocker.MagicMock(name="Scenario"))

    # then
    with pytest.raises(StepImplementationPatternNotSupported) as excinfo:
        # when
        match_step(step, registry_mock)

    # then
    assert excinfo.value.step_impl == step_impl


def test_matcher_should_match_step_impl_with_parse_type_pattern(mocker):
    """The matcher should match a Step with a parse-type pattern"""
    # given
    step_impl = StepImpl("Given", "pattern", None)
    registry_mock = mocker.MagicMock()
    registry_mock.step_implementations.return_value = [step_impl]
    step = Step(1, "Given", "Given", "pattern", None, None, None, None)
    step.set_scenario(mocker.MagicMock(name="Scenario"))

    # when
    match_step(step, registry_mock)

    # then
    assert step.step_impl == step_impl
    assert isinstance(step.step_impl_match, ParseTypeStepImplMatcher.Match)


def test_matcher_should_match_step_impl_with_regex_pattern(mocker):
    """The matcher should match a Step with a parse-type pattern"""
    # given
    step_impl = StepImpl("Given", re.compile(r"pattern"), None)
    registry_mock = mocker.MagicMock()
    registry_mock.step_implementations.return_value = [step_impl]
    step = Step(1, "Given", "Given", "pattern", None, None, None, None)
    step.set_scenario(mocker.MagicMock(name="Scenario"))

    # when
    match_step(step, registry_mock)

    # then
    assert step.step_impl == step_impl
    assert isinstance(step.step_impl_match, RegexStepImplMatcher.Match)


def test_matcher_should_match_best_step_impl_candidate(mocker):
    """The matcher should match the best matching Step Implementation Candidate"""
    # given
    step_impl_candidate_1 = StepImpl("Given", re.compile(r"fooo"), None)
    step_impl_candidate_2 = StepImpl("Given", re.compile(r"foo"), None)
    step_impl_candidate_3 = StepImpl("Given", re.compile(r"foooo"), None)
    step_impl_no_candidate = StepImpl("Given", re.compile(r"meh"), None)

    registry_mock = mocker.MagicMock()
    registry_mock.step_implementations.return_value = [
        step_impl_candidate_1,
        step_impl_candidate_2,
        step_impl_candidate_3,
        step_impl_no_candidate,
    ]
    step = Step(1, "Given", "Given", "foo", None, None, None, None)
    step.set_scenario(mocker.MagicMock(name="Scenario"))

    # when
    match_step(step, registry_mock)

    # then
    assert step.step_impl == step_impl_candidate_2


def test_matcher_should_match_step_impl_with_step_with_constants(mocker):
    """The matcher should match a Step with Constants"""
    # given
    step_impl = StepImpl("Given", "pattern with A and B", None)
    registry_mock = mocker.MagicMock()
    registry_mock.step_implementations.return_value = [step_impl]
    step = Step(
        1, "Given", "Given", "pattern with ${x} and ${y}", None, None, None, None
    )
    scenario_mock = mocker.MagicMock(name="Scenario")
    scenario_mock.constants = {"x": "A", "y": "B"}
    step.set_scenario(scenario_mock)

    # when
    match_step(step, registry_mock)

    # then
    assert step.step_impl == step_impl
    assert isinstance(step.step_impl_match, ParseTypeStepImplMatcher.Match)



import functools
import re
from pathlib import Path

from setuptools import find_packages, setup

#: Holds a list of packages to install with the binary distribution
PACKAGES = find_packages(where="src")
META_FILE = Path("src").absolute() / "radish" / "__init__.py"
KEYWORDS = ["testing", "bdd", "tdd", "gherkin", "cucumber", "automated"]
CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Environment :: Console",
    "Intended Audience :: Developers",
    "Intended Audience :: Education",
    "Intended Audience :: Other Audience",
    "License :: OSI Approved :: MIT License",
    "Natural Language :: English",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Education :: Testing",
    "Topic :: Software Development",
    "Topic :: Software Development :: Testing",
]

#: Holds the runtime requirements for the end user
INSTALL_REQUIRES = [
    "lark-parser==0.7.6a1",
    "click",
    "colorful>=0.5.4",
    "tag-expressions>=1.0.0",
    "parse_type>0.4.0",
    "humanize",
    "PyYAML",
]
#: Holds runtime requirements and development requirements
EXTRAS_REQUIRES = {
    # extras for end users
    "xml": ["lxml"],
    "ipython-debugger": ["ipython"],
    "coverage": ["coverage"],
    # extras for contributors
    "docs": ["sphinx", "towncrier"],
    "tests": ["freezegun", "coverage[toml]", "pytest", "pytest-mock"]
    + ["lxml", "PyYAML"],
}
EXTRAS_REQUIRES["dev"] = (
    EXTRAS_REQUIRES["tests"] + EXTRAS_REQUIRES["docs"] + ["pre-commit"]
)

#: Holds the contents of the README file
with open("README.md", encoding="utf-8") as readme:
    __README_CONTENTS__ = readme.read()


@functools.lru_cache()
def read(metafile):
    """
    Return the contents of the given meta data file assuming UTF-8 encoding.
    """
    with open(str(metafile), encoding="utf-8") as f:
        return f.read()


def get_meta(meta, metafile):
    """
    Extract __*meta*__ from the given metafile.
    """
    contents = read(metafile)
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta), contents, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


setup(
    name="radish-bdd",
    version=get_meta("version", META_FILE),
    license=get_meta("license", META_FILE),
    description=get_meta("description", META_FILE),
    long_description=__README_CONTENTS__,
    long_description_content_type="text/markdown",
    author=get_meta("author", META_FILE),
    author_email=get_meta("author_email", META_FILE),
    maintainer=get_meta("author", META_FILE),
    maintainer_email=get_meta("author_email", META_FILE),
    platforms=["Linux", "Windows", "MAC OS X"],
    url=get_meta("url", META_FILE),
    download_url=get_meta("download_url", META_FILE),
    bugtrack_url=get_meta("bugtrack_url", META_FILE),
    packages=PACKAGES,
    package_dir={"": "src"},
    package_data={"": ["radish/parser/grammer.g"]},
    include_package_data=True,
    python_requires=">=3.5.*",
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRES,
    entry_points={
        "console_scripts": [
            "radish = radish.__main__:cli",
            "radish-test = radish.step_testing.__main__:cli",
            "radish-parser = radish.parser.__main__:cli",
        ]
    },
    keywords=KEYWORDS,
    classifiers=CLASSIFIERS,
)



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import subprocess


import pytest


@pytest.fixture(name="radish_runner")
def setup_radish_runner(tmpdir):
    """Fixture to setup a radish command line runner"""

    def __runner(features, basedir, radish_args):
        cmdline = ["radish", *features, "-b", basedir, *radish_args]
        proc = subprocess.Popen(
            cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=str(tmpdir)
        )
        stdout, stderr = proc.communicate()
        return proc.returncode, stdout, stderr

    return __runner



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from radish.errors import RadishError
from radish.models.stepfailurereport import StepFailureReport


def test_stepfailurereport_model_should_correctly_analyse_a_exception():
    """The StepFailureReport model should correctly analyse a given Exception"""
    # given
    def raise_exc():
        raise RadishError("buuh!")

    raised_exception = None
    try:
        raise_exc()
    except RadishError as exc:
        # when
        report = StepFailureReport(exc)
        raised_exception = exc

    # then
    assert report.exception is raised_exception
    assert report.reason == "buuh!"
    assert report.name == "RadishError"



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from radish.models.scenario import Scenario


class Background(Scenario):
    """Represents a single instance of a Gherkin Background"""

    def __init__(
        self, keyword: str, short_description: str, path: str, line: int, steps
    ) -> None:
        super().__init__(0, keyword, short_description, [], path, line, steps)

        #: Holds the Scenario for which this Background is supposed to run.
        self.scenario = None

    def __repr__(self) -> str:
        return "<Background: '{short_description} with {steps} Steps @ {path}:{line}>".format(
            short_description=self.short_description,
            steps=len(self.steps),
            path=self.path,
            line=self.line,
        )  # pragma: no cover

    def set_scenario(self, scenario):
        """Set the Scenario instance for this Background

        Eventually each Background is run as part of a Scenario.
        The given Scenario will be used to access the execution hierarchy
        and other context sensitive functionalities.
        """
        self.scenario = scenario  # pragma: no cover
        self.context = self.scenario.context

    @property
    def state(self):
        """Get the State of this Scenario"""
        return self._steps_state(self.steps)



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import pydoc
import warnings
from collections import OrderedDict

import yaml


def get_debugger():
    """Return a debugger to debug Steps

    The debugger is evaluated according to the following rules:
    1. Check if IPython is installed, if so, use the IPython Pdb
    2. Use Python's built-in Pdb
    """
    try:
        from IPython.core.debugger import Pdb  # pragma: no cover

        pdb = Pdb()  # pragma: no cover
    except ImportError:
        warnings.warn(
            UserWarning(
                "Python's built-in pdb was selected as a debugger. "
                "If you want to use IPython as a debugger you have to "
                "'pip install radish-bdd[ipython-debugger]'"
            )
        )
        import pdb

    return pdb


def get_func_pos_args_as_kwargs(func, pos_arg_values):
    """Get the positional function arguments as keyword arguments given their values"""
    pos_arg_names = func.__code__.co_varnames[
        1 : func.__code__.co_argcount
    ]  # without the `step` argument
    kwargs = dict(zip(pos_arg_names, pos_arg_values))
    return kwargs


def locate_python_object(name):
    """
    Locate the object for the given name
    """
    obj = pydoc.locate(name)
    if not obj:
        obj = globals().get(name, None)

    return obj


def yaml_ordered_load(
    stream, loader_type=yaml.SafeLoader, object_pairs_hook=OrderedDict
):
    """Load YAML using an OrderedDict

    This function is needed for reproducibility with Python 3.5,
    which doesn't guarantee dict key ordering.
    """

    class OrderedLoader(loader_type):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping
    )

    return yaml.load(stream, OrderedLoader)



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""


class Context:
    """Represents a Context object for arbitrary data

    A ``Context`` object can be used by other models to
    store arbirtrary data during it's lifetime.
    """

    def __init__(self):
        self.constants = []



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from radish.models.tag import Tag


class PreconditionTag(Tag):
    """Represents a single radish Precondition Tag"""

    def __init__(
        self,
        feature_filename: str,
        scenario_short_description: str,
        path: str,
        line: int,
    ) -> None:
        super().__init__(
            "precondition({}: {})".format(feature_filename, scenario_short_description),
            path,
            line,
        )
        self.feature_filename = feature_filename
        self.scenario_short_description = scenario_short_description

    def __repr__(self) -> str:
        return "<PreconditionTag: {name} @ {path}:{line}>".format(
            name=self.name, path=self.path, line=self.line
        )  # pragma: no cover



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import pytest

from radish.loader import load_modules, load_module
from radish.errors import RadishError


def test_loader_should_raise_if_location_to_load_not_exists(mocker):
    """The Loader should raise an Exception if a location does not exist which should be loaded"""
    # given
    non_existant_location = mocker.MagicMock()
    non_existant_location.exists.return_value = False
    locations = [mocker.MagicMock(), non_existant_location]

    # then
    with pytest.raises(RadishError):
        # when
        load_modules(locations)


def test_loader_should_load_python_module_in_location(mocker):
    """The Loader should attempt to load Python module within a location"""
    # given
    load_module_mock = mocker.patch("radish.loader.load_module")
    module_mock = mocker.MagicMock(name="Python Module")
    location = mocker.MagicMock()
    location.glob.return_value = [module_mock]

    # when
    loaded_locations = load_modules([location])

    # then
    load_module_mock.assert_called_once_with(module_mock)
    assert loaded_locations == [module_mock]


def test_loader_should_fail_when_import_fails(mocker):
    """The Loader should fail if the import of a module fails"""
    # given
    spec_mock = mocker.patch("importlib.util.spec_from_file_location")
    spec_mock.return_value.loader.exec_module.side_effect = Exception("buuhu!")
    mocker.patch("importlib.util.module_from_spec")
    location = mocker.MagicMock("Python Module Path")
    location.stem = "foo"

    # then
    with pytest.raises(ImportError, match="Unable to import module 'foo' .*: buuhu!"):
        # when
        load_module(location)



"""
radish
~~~~~~

the root from red to green.  BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import textwrap

from radish.terrain import world


class RadishError(Exception):
    """Base-Exception for all radish based errors."""

    pass


class StepImplementationNotFoundError(RadishError):
    """Exception raised when no Step Implementation can be found for a Step"""

    def __init__(self, step):
        self.step = step

    def __str__(self) -> str:
        return textwrap.dedent(
            """
            Radish is unable to find a matching Step Implementation for the Step:
                "{keyword} {text}"

            This Step is defined in the Feature File {path} on Line {line}.
            You can register a Step Implementation by using the
            "{keyword_lower}" decorator and placing the code below in one of these modules:
            {basedirs}

            from radish import {keyword_lower}

            @{keyword_lower}("{text}")
            def my_awesome_step(step):
                raise NotImplementedError("Oups, no implementation for this Step, yet")
        """.format(
                keyword=self.step.keyword,
                keyword_lower=self.step.keyword.lower(),
                text=self.step.text,
                path=self.step.path,
                line=self.step.line,
                basedirs="\n".join(str(b) for b in world.config.basedirs),
            )
        )


class StepImplementationPatternNotSupported(RadishError):
    """
    Exception raised when a registered Step Implementation Pattern is not supported by any matcher
    """

    def __init__(self, step_impl):
        self.step_impl = step_impl


class StepBehaveLikeRecursionError(RadishError):
    """Exception raised when a recursion is detected in behave-like calls of a Step"""

    def __init__(self):
        super().__init__(
            "Detected a infinit recursion in your ``step.behave_like`` calls"
        )


class HookExecError(RadishError):
    """Exception raised when a Hook execution raised an Exception"""

    def __init__(self, hook_impl, orig_exc):
        self.hook_impl = hook_impl
        self.orig_exc = orig_exc
        super().__init__(
            "The '@{when}.{what}' Hook '{func_name}' raised an {exc_type}: {exc_msg}".format(
                when=hook_impl.when,
                what=hook_impl.what,
                func_name=hook_impl.func.__name__,
                exc_type=orig_exc.__class__.__name__,
                exc_msg=str(orig_exc),
            )
        )



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import pytest

from radish.models import Feature, State
from radish.models import ConstantTag, Tag


@pytest.mark.parametrize(
    "given_rule_states, expected_state",
    [
        ([State.PASSED, State.PASSED, State.PASSED], State.PASSED),
        ([State.PASSED, State.UNTESTED, State.PASSED], State.UNTESTED),
        ([State.PASSED, State.SKIPPED, State.UNTESTED], State.SKIPPED),
        ([State.PASSED, State.PENDING, State.SKIPPED], State.PENDING),
        ([State.PASSED, State.PENDING, State.FAILED], State.FAILED),
        ([State.PASSED, State.RUNNING, State.FAILED], State.RUNNING),
    ],
    ids=[
        "[State.PASSED, State.PASSED, State.PASSED] -> State.PASSED",
        "[State.PASSED, State.UNTESTED, State.PASSED] -> State.UNTESTED",
        "[State.PASSED, State.SKIPPED, State.UNTESTED] -> State.SKIPPED",
        "[State.PASSED, State.PENDING, State.SKIPPED] -> State.PENDING",
        "[State.PASSED, State.PENDING, State.FAILED] -> State.FAILED",
        "[State.PASSED, State.RUNNING, State.FAILED] -> State.RUNNING",
    ],
)
def test_feature_should_return_correct_state_according_to_its_rule_states(
    given_rule_states, expected_state, mocker
):
    """A Feature should return the correct State according to its rule States"""
    # given
    feature = Feature(
        1,
        "Feature",
        "My Feature",
        None,
        [],
        None,
        None,
        None,
        [mocker.MagicMock(state=s) for s in given_rule_states],
        None,
    )

    # when
    actual_state = feature.state

    # then
    assert actual_state == expected_state


@pytest.mark.parametrize(
    "rules_need_to_run, expected_has_to_run",
    [
        ([False, False, False], False),
        ([False, True, False], True),
        ([False, True, True], True),
    ],
    ids=["no Rule needs to run", "a Rule needs to run", "multiple Rules need to run"],
)
def test_feature_should_run_if_one_of_its_rule_has_to_run(
    rules_need_to_run, expected_has_to_run, mocker
):
    """A Feature should run if one of its Rules has to run"""
    rules = []
    for has_to_run in rules_need_to_run:
        rule_mock = mocker.MagicMock()
        rule_mock.has_to_run.return_value = has_to_run
        rules.append(rule_mock)

    # given
    feature = Feature(
        1, "Feature", "My Feature", None, [], None, None, None, rules, None
    )

    # when
    actual_has_to_run = feature.has_to_run(None, None)

    # then
    assert actual_has_to_run == expected_has_to_run


def test_feature_should_return_all_constants():
    """A Feature should return all Constants from the Tags"""
    # given
    tags = [
        Tag("x", None, None),
        ConstantTag("k1", "v1", None, None),
        Tag("y", None, None),
        ConstantTag("k2", "v2", None, None),
    ]
    feature = Feature(
        1, "Feature", "My Feature", None, tags, None, None, None, None, None
    )

    # when
    actual_constants = feature.constants

    # then
    assert actual_constants == {"k1": "v1", "k2": "v2"}



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import pytest

from radish.config import Config
from radish.errors import RadishError


def test_config_adds_items_as_attributes():
    """The Config objects adds its Config Items as object attributes"""
    # when
    config = Config({"foo": "bar", "bla": "meh"})

    # then
    assert config.foo == "bar"
    assert config.bla == "meh"


def test_config_parses_tag_expressions():
    """The Config object parses Tag Expressions"""
    # when
    config = Config({"tags": "foo and bar"})

    # then
    assert config.tag_expression is not None


def test_config_raise_error_for_invalid_tag_expressions():
    """The Config object should raise an Error for an invalid Tag Expression"""
    # then
    with pytest.raises(RadishError):
        # when
        Config({"tags": "foo and"})



"""
radish
~~~~~~

the root from red to green.  BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import copy

from radish.models.constant_tag import ConstantTag
from radish.models.context import Context
from radish.models.state import State
from radish.models.timed import Timed


class Scenario(Timed):
    """Represents a single instance of a Gherkin Scenario"""

    def __init__(
        self,
        scenario_id: int,
        keyword: str,
        short_description: str,
        tags,
        path: str,
        line: int,
        steps,
    ) -> None:
        super().__init__()
        self.id = scenario_id
        self.keyword = keyword
        self.short_description = short_description
        self.tags = tags
        self.path = path
        self.line = line
        self.steps = steps

        for step in self.steps:
            step.set_scenario(self)

        self.feature = None
        self.background = None
        self.rule = None
        self.preconditions = None

        #: Holds the ``Context`` object.
        self.context = Context()

    def __repr__(self) -> str:
        return "<Scenario: {id} '{short_description} with {steps} Steps @ {path}:{line}>".format(
            id=self.id,
            short_description=self.short_description,
            steps=len(self.steps),
            path=self.path,
            line=self.line,
        )  # pragma: no cover

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        return (
            self.short_description == other.short_description
            and self.path == other.path
            and self.line == other.line
        )

    def __hash__(self):
        return hash((self.short_description, self.path, self.line))

    def set_feature(self, feature):
        """Set the Feature for this Scenario"""
        self.feature = feature
        for step in self.steps:
            step.set_feature(feature)

    def set_background(self, background):
        """Set the Background for this Scenario"""
        if background:
            self.background = copy.deepcopy(background)
            self.background.set_scenario(self)
            self.background.set_rule(self.rule)
            self.background.set_feature(self.feature)

    def set_rule(self, rule):
        """Set the Rule for this Scenario"""
        self.rule = rule
        for step in self.steps:
            step.set_rule(rule)
        if self.preconditions:
            for precondition in self.preconditions:
                precondition.set_rule(rule)

    def set_preconditions(self, preconditions):
        """Set Preconditions for this Scenario"""
        if preconditions:
            self.preconditions = copy.deepcopy(preconditions)
            for precondition in self.preconditions:
                for step in precondition.steps:
                    step.set_scenario(self)

    def _steps_state(self, steps):
        """Get the State for some Steps"""
        return State.report_state(s.state for s in steps)

    @property
    def state(self):
        """Get the State of this Scenario"""
        steps = self.steps
        if self.background:
            state = self.background.state
            if state is State.FAILED:
                return state
            steps = self.steps + self.background.steps

        return self._steps_state(steps)

    @property
    def constants(self):
        """Get the Constants of this Scenario

        The Constants are lazy-evaluated from the Tags.
        """
        consts = self.feature.constants
        consts.update({t.key: t.value for t in self.tags if isinstance(t, ConstantTag)})
        return consts

    def get_all_tags(self):
        """Return all Tags for this Scenario

        These Tags include the ones inherited from the Feature, too.
        Use `Scenario.tags` to get only the tags declared for this Scenario.
        """
        return self.tags + self.feature.tags

    def has_to_run(self, tag_expression, scenario_ids):
        """Evaluate if this Scenario has to run or not

        The Scenario needs to be run if any of the following conditions is True:
        * No ``tag_expression`` is given
        * The ``tag_expression`` evaluates to True on the Scenario Tags
        * The ``tag_expression`` evaluates to True on the Feature Tags
        * No ``scenario_ids`` is given
        * The ``scenario_ids`` is given and contains this Scenarios Id
        """
        has_to_run_tag_expression = True
        has_to_run_scenario_ids = True

        if tag_expression:
            tag_names = [t.name for t in self.get_all_tags()]
            has_to_run_tag_expression = tag_expression.evaluate(tag_names)

        if scenario_ids:
            has_to_run_scenario_ids = self.id in scenario_ids

        if tag_expression and scenario_ids:
            return has_to_run_tag_expression and has_to_run_scenario_ids
        elif tag_expression:
            return has_to_run_tag_expression
        elif scenario_ids:
            return has_to_run_scenario_ids
        else:
            return True
